# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the class to connect to the Service Registry contract."""

import hashlib
import json
import logging
from pathlib import Path
from typing import Any, Dict, FrozenSet, List, Optional, Tuple, Union, cast

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi
from web3.types import BlockData, EventData, TxReceipt


PUBLIC_ID = PublicId.from_str("valory/service_registry:0.1.0")
ETHEREUM_IDENTIFIER = "ethereum"
UNIT_HASH_PREFIX = "0x{metadata_hash}"
EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID = {
    1: "0x48b6af7B12C71f09e2fC8aF4855De4Ff54e775cA",
    5: "0x1cEe30D08943EB58EFF84DD1AB44a6ee6FEff63a",
    100: "0x9338b5153AE39BB89f50468E608eD9d764B755fD",
    137: "0xE3607b00E75f6405248323A9417ff6b39B244b50",
    31337: "0x998abeb3E57409262aE5b751f60747921B33613E",
}
DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID = {
    1: "6f9fc7f3c2348801737120e6b5f8fa8e9670c65152c66d128ff4cddb465b4d705340c559e352f5e7f29bda3b84a8d36d4a9448b791cfe2d370e31c01276e0244",
    5: "d4a860f21f17762c99d93359244b39a878dd5bac9ea6056c0ff29c7558d6653aa0d5962aa819fc9f05f237d068845125cfc37a7fd7761b11c29a709ad5c48157",
    100: "10e2cfb500481d6c5a3b6b90507e4ac04d8b0d88741cea5568306ed4115f24e8b9747055423da5fca05838d5ccefebf41fb47d2ba1fb45215b6b21c0a27823be",
    137: "10e2cfb500481d6c5a3b6b90507e4ac04d8b0d88741cea5568306ed4115f24e8b9747055423da5fca05838d5ccefebf41fb47d2ba1fb45215b6b21c0a27823be",
    31337: "41ab54d43fd4bdfdc929658a0dc9bedd970c7339eecafbefda9892ab54c02c396bceedaa3a84a0f4690bee03dc11195a3267a264de7859c420efbf4291f1fef0",
}
L1_CHAINS = (
    1,
    5,
    31337,
)
L2_BUILD_FILENAME = "ServiceRegistryL2.json"

ServiceInfo = Tuple[int, str, bytes, int, int, int, int, List[int]]

_logger = logging.getLogger(
    f"aea.packages.{PUBLIC_ID.author}.contracts.{PUBLIC_ID.name}.contract"
)


class ServiceRegistryContract(Contract):
    """The Service Registry contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get the Safe transaction."""
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[bytes]:
        """Get raw message."""
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get state."""
        raise NotImplementedError  # pragma: nocover

    @staticmethod
    def is_l1_chain(ledger_api: LedgerApi) -> bool:
        """Check if we're interecting with an L1 chain"""
        return ledger_api.api.eth.chain_id in L1_CHAINS

    @staticmethod
    def load_l2_build() -> JSONLike:
        """Load L2 ABI"""
        path = Path(__file__).parent / "build" / L2_BUILD_FILENAME
        return json.loads(path.read_text(encoding="utf-8"))

    @classmethod
    def get_instance(
        cls,
        ledger_api: LedgerApi,
        contract_address: Optional[str] = None,
    ) -> Any:
        """Get contract instance."""
        if ledger_api.identifier != ETHEREUM_IDENTIFIER:
            return super().get_instance(
                ledger_api=ledger_api, contract_address=contract_address
            )
        if cls.is_l1_chain(ledger_api=ledger_api):
            contract_interface = cls.contract_interface.get(ledger_api.identifier, {})
        else:
            contract_interface = cls.load_l2_build()
        instance = ledger_api.get_contract_instance(
            contract_interface, contract_address
        )
        return instance

    @classmethod
    def verify_contract(
        cls, ledger_api: LedgerApi, contract_address: str
    ) -> Dict[str, Union[bool, str]]:
        """
        Verify the contract's bytecode

        :param ledger_api: the ledger API object
        :param contract_address: the contract address
        :return: the verified status
        """
        verified = False
        chain_id = ledger_api.api.eth.chain_id
        expected_address = EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID[chain_id]
        if contract_address != expected_address:
            _logger.error(
                f"For chain_id {chain_id} expected {expected_address} and got {contract_address}."
            )
            return dict(verified=verified)
        deployed_bytecode = ledger_api.api.eth.get_code(contract_address).hex()
        sha512_hash = hashlib.sha512(deployed_bytecode.encode("utf-8")).hexdigest()
        verified = DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID[chain_id] == sha512_hash
        if not verified:  # pragma: nocover
            _logger.error(
                f"CONTRACT NOT VERIFIED! Contract address: {contract_address}, chain_id: {chain_id}."
            )
        return dict(verified=verified)

    @classmethod
    def exists(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> bool:
        """Check if the service id exists"""

        contract_instance = cls.get_instance(ledger_api, contract_address)
        exists = ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="exists",
            unitId=service_id,
        )

        return cast(bool, exists)

    @classmethod
    def get_agent_instances(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> Dict[str, Any]:
        """Retrieve on-chain agent instances."""

        contract_instance = cls.get_instance(ledger_api, contract_address)
        service_info = ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="getAgentInstances",
            serviceId=service_id,
        )

        return dict(
            numAgentInstances=service_info[0],
            agentInstances=service_info[1],
        )

    @classmethod
    def get_service_owner(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> Dict[str, Any]:
        """Retrieve the service owner."""
        contract_instance = cls.get_instance(ledger_api, contract_address)
        service_owner = contract_instance.functions.ownerOf(service_id).call()
        checksum_service_owner = ledger_api.api.to_checksum_address(service_owner)
        return dict(
            service_owner=checksum_service_owner,
        )

    @classmethod
    def get_service_information(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        token_id: int,
    ) -> ServiceInfo:
        """Retrieve service information"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.functions.getService(token_id).call()

    @classmethod
    def get_token_uri(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        token_id: int,
    ) -> str:
        """Returns the latest metadata URI for a component."""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.functions.tokenURI(token_id).call()

    @classmethod
    def get_create_events(  # pragma: nocover
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        receipt: JSONLike,
    ) -> Optional[int]:
        """Returns `CreateUnit` event filter."""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.events.CreateService().process_receipt(receipt)

    @classmethod
    def get_update_events(  # pragma: nocover
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        receipt: JSONLike,
    ) -> Optional[int]:
        """Returns `CreateUnit` event filter."""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.events.UpdateService().process_receipt(receipt)

    @classmethod
    def get_update_hash_events(  # pragma: nocover
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        receipt: JSONLike,
    ) -> Optional[int]:
        """Returns `CreateUnit` event filter."""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.events.UpdateUnitHash().process_receipt(receipt)

    @classmethod
    def get_events(  # pragma: nocover
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        event: str,
        receipt: JSONLike,
    ) -> Dict:
        """Process receipt for events."""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        Event = getattr(contract_interface.events, event, None)
        if Event is None:
            return {"events": []}
        return {"events": Event().process_receipt(receipt)}

    @classmethod
    def process_receipt(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        event: str,
        receipt: JSONLike,
    ) -> JSONLike:
        """Checks for a successful service registration event in the latest block"""
        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        Event = getattr(contract_interface.events, event, None)
        if Event is None:
            return {"events": []}
        return {"events": Event().process_receipt(receipt)}

    @classmethod
    def get_slash_data(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        agent_instances: List[str],
        amounts: List[int],
        service_id: int,
    ) -> Dict[str, bytes]:
        """Gets the encoded arguments for a slashing tx, which should only be called via the multisig."""

        contract_instance = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )

        slash_kwargs = dict(
            agentInstances=agent_instances,
            amounts=amounts,
            serviceId=service_id,
        )

        data = contract_instance.encodeABI(fn_name="slash", kwargs=slash_kwargs)
        return {"data": bytes.fromhex(data[2:])}

    @classmethod
    def process_slash_receipt(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        tx_hash: str,
    ) -> Optional[JSONLike]:
        """
        Process the slash receipt.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param tx_hash: the hash of a slash tx to be processed.
        :return: a dictionary with the timestamp of the slashing and the `OperatorSlashed` events.
        """
        contract = cls.get_instance(ledger_api, contract_address)
        receipt: TxReceipt = ledger_api.api.eth.get_transaction_receipt(tx_hash)
        logs: List[EventData] = list(
            contract.events.OperatorSlashed().process_receipt(receipt)
        )

        if len(logs) == 0:
            _logger.error(f"No `OperatorSlashed` event was emitted in tx {tx_hash}.")
            return None

        block: BlockData = ledger_api.api.eth.get_block(receipt["blockNumber"])

        return {
            "slash_timestamp": block["timestamp"],
            "events": [log["args"] for log in logs],
        }

    @classmethod
    def _get_operator(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        agent_instance: str,
    ) -> str:
        """Retrieve an operator given its agent instance."""

        contract_instance = cls.get_instance(ledger_api, contract_address)
        map_fn = contract_instance.functions.mapAgentInstanceOperators
        return map_fn(agent_instance).call()

    @classmethod
    def get_operators_mapping(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        agent_instances: FrozenSet[str],
    ) -> Dict[str, str]:
        """
        Retrieve a mapping of the given agent instances to their operators.

        Please keep in mind that this method performs a call for each agent instance.

        :param ledger_api: the ledger api.
        :param contract_address: the contract address.
        :param agent_instances: the agent instances to be mapped.
        :return: a mapping of the given agent instances to their operators.
        """
        return {
            agent: cls._get_operator(ledger_api, contract_address, agent)
            for agent in agent_instances
        }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for the slashing background skill."""

from dataclasses import dataclass, field

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


@dataclass(frozen=True)
class SlashingTxPayload(BaseTxPayload):
    """Represent a transaction payload for slashing."""

    # normal payload field
    tx_hex: str
    # these two fields are present in order to simplify the `end_block` method implementation of the corresponding round
    in_progress: bool = field(default=True)
    sent: bool = field(default=True)


@dataclass(frozen=True)
class StatusResetPayload(BaseTxPayload):
    """Represent a transaction payload for resetting the offence status."""

    # normal payload fields
    operators_mapping: str
    slash_timestamps: str
    # these two fields are present in order to simplify the `end_block` method implementation of the corresponding round
    slashing_in_flight: bool = field(default=False)
    slash_tx_sent: bool = field(default=False)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `mint` command group."""

from typing import Any
from unittest import mock

import pytest
from aea.configurations.data_types import PackageId
from aea_test_autonomy.configurations import ETHEREUM_KEY_DEPLOYER
from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401

from autonomy.chain.config import ChainConfigs, ChainType
from autonomy.chain.service import get_service_info
from autonomy.cli.helpers.chain import OnChainHelper

from tests.test_autonomy.test_chain.base import (
    BaseChainInteractionTest,
    DEFAULT_SERVICE_MINT_PARAMETERS,
    DUMMY_AGENT,
    DUMMY_CONNECTION,
    DUMMY_CONTRACT,
    DUMMY_PACKAGE_MANAGER,
    DUMMY_PROTOCOL,
    DUMMY_SERVICE,
    DUMMY_SKILL,
    patch_component_verification,
    patch_subgraph,
)
from tests.test_autonomy.test_cli.base import BaseCliTest


CUSTOM_OWNER = "0x8626F6940E2EB28930EFB4CEF49B2D1F2C9C1199"


class DummyContract:
    """Dummy contract"""

    def get_create_events(self, *args: Any, **kwargs: Any) -> None:
        """Dummy method implementation"""

    def get_create_transaction(self, *args: Any, **kwargs: Any) -> None:
        """Dummy method implementation"""


@patch_component_verification()
class TestMintComponents(BaseChainInteractionTest):
    """Test `autonomy develop mint` command."""

    cli_options = ("mint",)

    @pytest.mark.parametrize(
        argnames=("package_id"),
        argvalues=(
            DUMMY_PROTOCOL,
            DUMMY_CONTRACT,
            DUMMY_CONNECTION,
            DUMMY_SKILL,
            DUMMY_AGENT,
        ),
    )
    def test_mint_components(self, package_id: PackageId) -> None:
        """Test mint components."""

        commands = [
            package_id.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=package_id
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
        ]

        result = self.run_cli(commands=tuple(commands))
        assert result.exit_code == 0, result.stderr
        assert "Component minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output
        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=package_id,
        )

        commands += ["--update", str(token_id)]
        result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 0, result.stderr
        assert "Component hash updated:" in result.output
        assert f"Token ID: {token_id}" in result.output
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_component_with_owner(
        self,
    ) -> None:
        """Test mint components."""

        commands = [
            DUMMY_PROTOCOL.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_PROTOCOL
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "--owner",
            CUSTOM_OWNER,
        ]

        result = self.run_cli(commands=tuple(commands))
        assert result.exit_code == 0, result.output
        assert "Component minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_owner_address(
            token_id=token_id,
            owner=CUSTOM_OWNER,
            package_id=DUMMY_PROTOCOL,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_service(
        self,
    ) -> None:
        """Test mint components."""
        agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])
        commands = [
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:],
        ]

        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 0, result
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=DUMMY_SERVICE,
        )

        commands += ["--update", str(token_id)]
        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))
        assert result.exit_code == 0, result.stderr
        assert "Service updated with:" in result.output
        assert f"Token ID: {token_id}" in result.output
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_service_without_threshold(
        self,
    ) -> None:
        """Test mint components."""
        agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])
        commands = [
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:-2],
        ]

        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 0, result.output
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=DUMMY_SERVICE,
        )
        ledger_api, _ = OnChainHelper.get_ledger_and_crypto_objects(
            chain_type=ChainType.LOCAL
        )
        _, _, _, threshold, *_ = get_service_info(
            ledger_api=ledger_api, chain_type=ChainType.LOCAL, token_id=token_id
        )
        assert threshold == 3

        # Test updates
        commands += ["--update", str(token_id)]
        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))
        assert result.exit_code == 0, result.stderr
        assert "Service updated with:" in result.output
        assert f"Token ID: {token_id}" in result.output

        self.verify_and_remove_metadata_file(token_id=token_id)

        ledger_api, _ = OnChainHelper.get_ledger_and_crypto_objects(
            chain_type=ChainType.LOCAL
        )
        _, _, _, threshold, *_ = get_service_info(
            ledger_api=ledger_api, chain_type=ChainType.LOCAL, token_id=token_id
        )
        assert threshold == 3

    def test_update_service_failure(
        self,
    ) -> None:
        """Test mint components."""
        agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])
        commands = [
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:],
        ]

        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 0, result
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_minted_token_id(
            token_id=token_id,
            package_id=DUMMY_SERVICE,
        )
        self.service_manager.activate(service_id=token_id)

        commands += ["--update", str(token_id)]
        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=tuple(commands))

        assert result.exit_code == 1, result.stdout
        assert (
            "Cannot update service hash, service needs to be in the pre-registration state"
            in result.stderr
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_mint_service_with_owner(
        self,
    ) -> None:
        """Test mint components."""
        agent_id = self.mint_component(package_id=DUMMY_AGENT, dependencies=[1])
        commands = (
            DUMMY_SERVICE.package_type.value,
            str(
                DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                    package_id=DUMMY_SERVICE
                )
            ),
            "--key",
            str(ETHEREUM_KEY_DEPLOYER),
            "-a",
            str(agent_id),
            *DEFAULT_SERVICE_MINT_PARAMETERS[2:],
            "--owner",
            CUSTOM_OWNER,
        )
        with patch_subgraph(
            response=[{"tokenId": agent_id, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(commands=commands)

        assert result.exit_code == 0, result
        assert "Service minted with:" in result.output
        assert "Metadata Hash:" in result.output
        assert "Token ID:" in result.output

        token_id = self.extract_token_id_from_output(output=result.output)
        self.verify_owner_address(
            token_id=token_id,
            owner=CUSTOM_OWNER,
            package_id=DUMMY_SERVICE,
        )
        self.verify_and_remove_metadata_file(token_id=token_id)

    def test_bad_owner_string(
        self,
    ) -> None:
        """Test connection error."""
        with patch_subgraph(
            response=[{"tokenId": 1, "publicId": "dummy_author/dummy_agent"}]
        ):
            result = self.run_cli(
                commands=(
                    "service",
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            package_id=DUMMY_SERVICE
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    *DEFAULT_SERVICE_MINT_PARAMETERS,
                    "--owner",
                    "0xowner",
                ),
            )
        self.cli_runner.mix_stderr = True
        assert result.exit_code == 1, result.output
        assert "Invalid owner address 0xowner" in result.stderr

    def test_fail_dependency_does_not_match_service(
        self,
    ) -> None:
        """Test token id retrieval failure."""
        with mock.patch.object(
            ChainConfigs, "get", return_value=ChainConfigs.local
        ), mock.patch(
            "autonomy.chain.utils.resolve_component_id",
            return_value={"name": "skill/author/name"},
        ):
            result = self.run_cli(
                commands=(
                    "--use-ethereum",
                    DUMMY_SERVICE.package_type.value,
                    str(
                        DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                            DUMMY_SERVICE
                        )
                    ),
                    "--key",
                    str(ETHEREUM_KEY_DEPLOYER),
                    "--nft",
                    "Qmbh9SQLbNRawh9Km3PMEDSxo77k1wib8fYZUdZkhPBiev",
                    *DEFAULT_SERVICE_MINT_PARAMETERS,
                ),
            )

        assert result.exit_code == 1, result.output
        assert (
            "Error: Public ID `valory/hello_world:any` for token 1 does not "
            "match with the one defained in the service package `dummy_author/dummy_agent:any`"
            in result.stderr
        )

    def test_dry_run(self) -> None:
        """Test dry run."""
        result = self.run_cli(
            commands=(
                "--dry-run",
                DUMMY_PROTOCOL.package_type.value,
                str(
                    DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                        package_id=DUMMY_PROTOCOL
                    )
                ),
                "--key",
                str(ETHEREUM_KEY_DEPLOYER),
            )
        )
        assert result.exit_code == 0, result.stderr
        assert "=== Dry run output ===" in result.output


class TestConnectionError(BaseCliTest):
    """Test connection error."""

    cli_options = ("mint",)

    def test_connection_error(self) -> None:
        """Test connection error."""
        self.cli_runner.mix_stderr = False
        result = self.run_cli(
            commands=(
                "protocol",
                str(
                    DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
                        package_id=DUMMY_PROTOCOL
                    )
                ),
                "--key",
                str(ETHEREUM_KEY_DEPLOYER),
            ),
        )
        assert result.exit_code == 1, result.output
        assert "RPCError(Cannot connect to the given RPC)" in result.stderr



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Localhost Deployment Generator."""
import json
import subprocess  # nosec
import typing as t
from pathlib import Path

from aea.configurations.constants import (
    DEFAULT_LEDGER,
    LEDGER,
    PRIVATE_KEY,
    PRIVATE_KEY_PATH_SCHEMA,
)

from autonomy.deploy.base import BaseDeploymentGenerator
from autonomy.deploy.constants import (
    AGENT_VARS_CONFIG_FILE,
    DEFAULT_ENCODING,
    TENDERMINT_VARS_CONFIG_FILE,
    TM_ENV_CREATE_EMPTY_BLOCKS,
    TM_ENV_P2P_LADDR,
    TM_ENV_PROXY_APP,
    TM_ENV_RPC_LADDR,
    TM_ENV_TMHOME,
    TM_ENV_TMSTATE,
    TM_ENV_USE_GRPC,
    TM_STATE_DIR,
)
from autonomy.deploy.generators.localhost.utils import (
    check_tendermint_version,
    setup_agent,
)


class HostDeploymentGenerator(BaseDeploymentGenerator):
    """Localhost deployment."""

    output_name: str = AGENT_VARS_CONFIG_FILE
    deployment_type: str = "localhost"

    @property
    def agent_dir(self) -> Path:
        """Path to the agent directory."""
        return self.build_dir / "agent"

    def generate_config_tendermint(self) -> "HostDeploymentGenerator":
        """Generate tendermint configuration."""
        tmhome = str(self.build_dir / "node")
        tendermint_executable = check_tendermint_version()
        # if executable found, setup its configs
        subprocess.run(  # pylint: disable=subprocess-run-check # nosec
            args=[
                tendermint_executable,
                "--home",
                tmhome,
                "init",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # TODO: Dynamic port allocation
        params = {
            TM_ENV_TMHOME: tmhome,
            TM_ENV_TMSTATE: str(self.build_dir / TM_STATE_DIR),
            TM_ENV_P2P_LADDR: "tcp://localhost:26656",
            TM_ENV_RPC_LADDR: "tcp://localhost:26657",
            TM_ENV_PROXY_APP: "tcp://localhost:26658",
            TM_ENV_CREATE_EMPTY_BLOCKS: "true",
            TM_ENV_USE_GRPC: "false",
        }
        (self.build_dir / TENDERMINT_VARS_CONFIG_FILE).write_text(
            json.dumps(params, indent=2),
        )

        return self

    def generate(
        self,
        image_version: t.Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "HostDeploymentGenerator":
        """Generate agent and tendermint configurations"""
        agent = self.service_builder.generate_agent(agent_n=0)
        self.output = json.dumps(
            {key: f"{value}" for key, value in agent.items()}, indent=2
        )
        (self.build_dir / self.output_name).write_text(
            json.dumps(self.output, indent=2),
        )

        return self

    def _populate_keys(self) -> None:
        """Populate the keys directory"""
        kp, *_ = t.cast(t.List[t.Dict[str, str]], self.service_builder.keys)
        key = kp[PRIVATE_KEY]
        ledger = kp.get(LEDGER, DEFAULT_LEDGER)
        keys_file = self.build_dir / PRIVATE_KEY_PATH_SCHEMA.format(ledger)
        keys_file.write_text(key, encoding=DEFAULT_ENCODING)
        setup_agent(self.build_dir, self.agent_dir, keys_file)

    def _populate_keys_multiledger(self) -> None:
        """Populate the keys directory with multiple set of keys"""

    def populate_private_keys(self) -> "HostDeploymentGenerator":
        """Populate the private keys to the build directory for host mapping."""
        if self.service_builder.multiledger:
            self._populate_keys_multiledger()
        else:
            self._populate_keys()
        return self

    def write_config(self) -> "BaseDeploymentGenerator":
        """Write output to build dir"""
        super().write_config()
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'test_solana_behaviour' skill."""
import json
from abc import ABC
from typing import Any, Dict, Generator, Optional, Set, Type, cast

from packages.valory.contracts.squads_multisig.contract import SquadsMultisig
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.registration_abci.behaviours import (
    AgentRegistrationRoundBehaviour,
    RegistrationStartupBehaviour,
)
from packages.valory.skills.reset_pause_abci.behaviours import (
    ResetPauseABCIConsensusBehaviour,
)
from packages.valory.skills.squads_transaction_settlement_abci.behaviours import (
    SolanaTransactionSettlementRoundBehaviour,
)
from packages.valory.skills.test_solana_tx_abci.composition import ComposedAbciApp
from packages.valory.skills.test_solana_tx_abci.models import Params
from packages.valory.skills.test_solana_tx_abci.payloads import SolanaTransactionPayload
from packages.valory.skills.test_solana_tx_abci.rounds import SolanaRound


DUMMY_ADDRESS = "0x0"
SOLANA = "solana"


class SolanaTransferBehaviour(BaseBehaviour, ABC):
    """A behaviour that makes a solana transfer via the squad multisig."""

    matching_round = SolanaRound

    @property
    def params(self) -> Params:
        """Get the params."""
        return cast(Params, self.context.params)

    def async_act(self) -> Generator:
        """Do the action."""
        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            payload_content = yield from self.get_payload_content()
            sender = self.context.agent_address
            payload = SolanaTransactionPayload(sender=sender, **payload_content)
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

    def _get_transfer_tx(self) -> Generator[None, None, Optional[Dict]]:
        """Get the transfer tx."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(SquadsMultisig.contract_id),
            contract_callable="get_transfer_tx",
            contract_address=DUMMY_ADDRESS,
            from_pubkey=self.params.squad_vault,
            to_pubkey=self.params.transfer_to_pubkey,
            lamports=self.params.transfer_lamports,
            chain_id=SOLANA,
            ledger_id=SOLANA,
        )
        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get the transfer tx. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        data = cast(Optional[Dict], response.state.body.get("data", None))
        return data

    def get_payload_content(self) -> Generator[None, None, Dict[str, Any]]:
        """Get the payload data."""
        transfer_tx = yield from self._get_transfer_tx()
        if transfer_tx is None:
            self.context.logger.error("Couldn't get the transfer tx.")
            return dict(instructions="", error=True)

        return dict(instructions=json.dumps([transfer_tx]), error=False)


class TestAbciConsensusBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the solana test abci app."""

    initial_behaviour_cls = RegistrationStartupBehaviour
    abci_app_cls = ComposedAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        *AgentRegistrationRoundBehaviour.behaviours,
        *ResetPauseABCIConsensusBehaviour.behaviours,
        *SolanaTransactionSettlementRoundBehaviour.behaviours,
        SolanaTransferBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/types/params.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2

from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1dtendermint/types/params.proto\x12\x10tendermint.types\x1a\x14gogoproto/gogo.proto\x1a\x1egoogle/protobuf/duration.proto"\xf3\x01\n\x0f\x43onsensusParams\x12\x32\n\x05\x62lock\x18\x01 \x01(\x0b\x32\x1d.tendermint.types.BlockParamsB\x04\xc8\xde\x1f\x00\x12\x38\n\x08\x65vidence\x18\x02 \x01(\x0b\x32 .tendermint.types.EvidenceParamsB\x04\xc8\xde\x1f\x00\x12:\n\tvalidator\x18\x03 \x01(\x0b\x32!.tendermint.types.ValidatorParamsB\x04\xc8\xde\x1f\x00\x12\x36\n\x07version\x18\x04 \x01(\x0b\x32\x1f.tendermint.types.VersionParamsB\x04\xc8\xde\x1f\x00"G\n\x0b\x42lockParams\x12\x11\n\tmax_bytes\x18\x01 \x01(\x03\x12\x0f\n\x07max_gas\x18\x02 \x01(\x03\x12\x14\n\x0ctime_iota_ms\x18\x03 \x01(\x03"~\n\x0e\x45videnceParams\x12\x1a\n\x12max_age_num_blocks\x18\x01 \x01(\x03\x12=\n\x10max_age_duration\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationB\x08\xc8\xde\x1f\x00\x98\xdf\x1f\x01\x12\x11\n\tmax_bytes\x18\x03 \x01(\x03"2\n\x0fValidatorParams\x12\x15\n\rpub_key_types\x18\x01 \x03(\t:\x08\xb8\xa0\x1f\x01\xe8\xa0\x1f\x01".\n\rVersionParams\x12\x13\n\x0b\x61pp_version\x18\x01 \x01(\x04:\x08\xb8\xa0\x1f\x01\xe8\xa0\x1f\x01">\n\x0cHashedParams\x12\x17\n\x0f\x62lock_max_bytes\x18\x01 \x01(\x03\x12\x15\n\rblock_max_gas\x18\x02 \x01(\x03\x42=Z7github.com/tendermint/tendermint/proto/tendermint/types\xa8\xe2\x1e\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "tendermint.types.params_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z7github.com/tendermint/tendermint/proto/tendermint/types\250\342\036\001"
    )
    _CONSENSUSPARAMS.fields_by_name["block"]._options = None
    _CONSENSUSPARAMS.fields_by_name["block"]._serialized_options = b"\310\336\037\000"
    _CONSENSUSPARAMS.fields_by_name["evidence"]._options = None
    _CONSENSUSPARAMS.fields_by_name[
        "evidence"
    ]._serialized_options = b"\310\336\037\000"
    _CONSENSUSPARAMS.fields_by_name["validator"]._options = None
    _CONSENSUSPARAMS.fields_by_name[
        "validator"
    ]._serialized_options = b"\310\336\037\000"
    _CONSENSUSPARAMS.fields_by_name["version"]._options = None
    _CONSENSUSPARAMS.fields_by_name["version"]._serialized_options = b"\310\336\037\000"
    _EVIDENCEPARAMS.fields_by_name["max_age_duration"]._options = None
    _EVIDENCEPARAMS.fields_by_name[
        "max_age_duration"
    ]._serialized_options = b"\310\336\037\000\230\337\037\001"
    _VALIDATORPARAMS._options = None
    _VALIDATORPARAMS._serialized_options = b"\270\240\037\001\350\240\037\001"
    _VERSIONPARAMS._options = None
    _VERSIONPARAMS._serialized_options = b"\270\240\037\001\350\240\037\001"
    _globals["_CONSENSUSPARAMS"]._serialized_start = 106
    _globals["_CONSENSUSPARAMS"]._serialized_end = 349
    _globals["_BLOCKPARAMS"]._serialized_start = 351
    _globals["_BLOCKPARAMS"]._serialized_end = 422
    _globals["_EVIDENCEPARAMS"]._serialized_start = 424
    _globals["_EVIDENCEPARAMS"]._serialized_end = 550
    _globals["_VALIDATORPARAMS"]._serialized_start = 552
    _globals["_VALIDATORPARAMS"]._serialized_end = 602
    _globals["_VERSIONPARAMS"]._serialized_start = 604
    _globals["_VERSIONPARAMS"]._serialized_end = 650
    _globals["_HASHEDPARAMS"]._serialized_start = 652
    _globals["_HASHEDPARAMS"]._serialized_end = 714
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the slashing background behaviours."""

import json
from abc import ABC
from calendar import timegm
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generator,
    List,
    Optional,
    Set,
    Type,
    cast,
)

from aea.protocols.base import Message

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.base import OffenceStatus, RoundSequence
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    AsyncBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.behaviours import AbstractRoundBehaviour
from packages.valory.skills.abstract_round_abci.utils import inverse
from packages.valory.skills.slashing_abci.composition import SlashingAbciApp
from packages.valory.skills.slashing_abci.models import SharedState
from packages.valory.skills.slashing_abci.payloads import (
    SlashingTxPayload,
    StatusResetPayload,
)
from packages.valory.skills.slashing_abci.rounds import (
    SlashingCheckRound,
    StatusResetRound,
)
from packages.valory.skills.slashing_abci.rounds import (
    SynchronizedData as SlashingSyncedData,
)
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    TransactionSettlementRoundBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    hash_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSyncedData,
)
from packages.valory.skills.transaction_settlement_abci.rounds import TX_HASH_LENGTH


# setting the safe gas to 0 means that all available gas will be used
# which is what we want in most cases
# more info here: https://safe-docs.dev.gnosisdev.com/safe/docs/contracts_tx_execution/
_SAFE_GAS = 0
# hardcoded to 0 because we don't need to send any ETH when slashing
_ETHER_VALUE = 0


class SlashingBaseBehaviour(BaseBehaviour, ABC):
    """Represents the base class for the slashing background FSM."""

    @property
    def shared_state(self) -> SharedState:
        """Get the round sequence from the shared state."""
        return cast(SharedState, self.context.state)

    @property
    def round_sequence(self) -> RoundSequence:
        """Get the round sequence from the shared state."""
        return self.shared_state.round_sequence

    @property
    def offence_status(self) -> Dict[str, OffenceStatus]:
        """Get the offence status from the round sequence."""
        return self.round_sequence.offence_status

    @offence_status.setter
    def offence_status(self, status: Dict[str, OffenceStatus]) -> None:
        """Set the offence status in the round sequence."""
        self.round_sequence.offence_status = status


class SlashingCheckBehaviour(SlashingBaseBehaviour):
    """
    A behaviour responsible for checking if there are any slashable events.

    Running concurrently with the other behaviours.
    """

    matching_round = SlashingCheckRound

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the slashing check behaviour."""
        super().__init__(**kwargs)
        self._slash_amounts: Dict[str, float] = {}

    @property
    def synchronized_data(self) -> SlashingSyncedData:
        """
        Return the synchronized data.

        Note: we instantiate here, rather than cast, as this runs
        concurrently and so the instantiation needs to happen somewhere.
        """
        return SlashingSyncedData(db=super().synchronized_data.db)

    @property
    def slashable_instances(self) -> List[str]:
        """Get the agent instances that are pending to be slashed."""
        return list(self._slash_amounts.keys())

    @property
    def slashable_amounts(self) -> List[float]:
        """Get the amounts that are pending to be slashed."""
        return list(self._slash_amounts.values())

    def is_majority_possible(self) -> bool:
        """Checks whether the service has enough participants to reach consensus."""
        return (
            self.synchronized_data.nb_participants
            >= self.synchronized_data.consensus_threshold
        )

    def is_slashing_majority_reached(self) -> bool:
        """Rely on the round to decide when the majority is reached."""
        return self.synchronized_data.slashing_majority_reached

    def is_round_ended(self, round_id: str) -> Callable[[], bool]:
        """
        Get a callable to check whether the current round has ended.

        We consider the background round ended when we have majority on the termination transaction.
        Overriden to allow for the behaviour to send transactions at any time.

        :return: the termination majority callable
        """
        return self.is_slashing_majority_reached

    def get_callback_request(self) -> Callable[[Message, "BaseBehaviour"], None]:
        """Wrapper for callback_request(), overridden to avoid mix-ups with normal (non-background) behaviours."""

        def callback_request(
            message: Message, _current_behaviour: BaseBehaviour
        ) -> None:
            """
            This method gets called when a response for a prior request is received.

            Overridden to remove the check that checks whether the behaviour that made the request is still active.
            The received message gets passed to the behaviour that invoked it, in this case it's always the
            background behaviour.

            :param message: the response.
            :param _current_behaviour: not used, left in to satisfy the interface.
            :return: none
            """
            if self.is_stopped:
                self.context.logger.info(
                    "dropping message as behaviour has stopped: %s", message
                )
                return

            if self.state != AsyncBehaviour.AsyncState.WAITING_MESSAGE:
                self.context.logger.warning(
                    f"could not send message {message} to {self.behaviour_id}"
                )
                return

            self.try_send(message)

        return callback_request

    def _check_offence_status(self) -> None:
        """Check the offence status, calculate the slash amount per operator, and assign it to `_slash_amounts`."""
        self._slash_amounts = {}

        for agent, status in self.offence_status.items():
            amount = status.slash_amount(
                self.params.light_slash_unit_amount,
                self.params.serious_slash_unit_amount,
            )
            # If an agent has not been slashed, then it is not included in the timestamps.
            # To facilitate the comparison in the subsequent code,
            # we assign a timestamp to the negative value of the wait time between slashes(`slash_cooldown_hours`).
            # This ensures that the comparison being performed is against 0.
            last_slashed_timestamp = self.synchronized_data.slash_timestamps.get(
                agent, -self.params.slash_cooldown_hours
            )
            last_round_transition_timestamp = timegm(
                self.round_sequence.last_round_transition_timestamp.utctimetuple()
            )

            if (
                amount > self.params.slash_threshold_amount
                and last_round_transition_timestamp
                > last_slashed_timestamp + self.params.slash_cooldown_hours
            ):
                # Check whether the threshold has been met, essentially bundling up offences.
                # Otherwise, we might end up in a situation where the slash tx is more expensive than the slash amount.
                # This in itself might be an attack vector, i.e., some malicious agent keeps sending bad payloads,
                # so that it drains the funds of the agents.
                # Moreover, only store amounts for agents which are not inside the slash wait window,
                # i.e., have not been slashed recently.
                self._slash_amounts[agent] = amount

    def _get_slash_data(self) -> Generator[None, None, Optional[bytes]]:
        """Get the slash tx data encoded."""
        response_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.params.service_registry_address,
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="get_slash_data",
            agent_instances=self.slashable_instances,
            amounts=self.slashable_amounts,
            service_id=self.params.on_chain_service_id,
            chain_id=self.params.default_chain_id,
        )
        if response_msg.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.error(
                f"Could not get the data for the slash transaction: {response_msg}"
            )
            return None

        slash_data = response_msg.raw_transaction.body.get("data", None)
        if slash_data is None:
            self.context.logger.error(
                "Something went wrong while trying to encode the slash data."
            )

        return slash_data

    def _get_safe_tx_hash(self, data: bytes) -> Generator[None, None, Optional[str]]:
        """
        Prepares and returns the safe tx hash.

        :param data: the safe tx data.
        :return: the tx hash
        """
        response_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_raw_safe_transaction_hash",
            to_address=self.params.service_registry_address,
            value=_ETHER_VALUE,
            data=data,
            safe_tx_gas=_SAFE_GAS,
            chain_id=self.params.default_chain_id,
        )

        if response_msg.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.error(
                "Couldn't get safe tx hash. Expected response performative "
                f"{ContractApiMessage.Performative.RAW_TRANSACTION.value}, received {response_msg}."  # type: ignore
            )
            return None

        tx_hash = response_msg.raw_transaction.body.get("tx_hash", None)
        if tx_hash is None or len(tx_hash) != TX_HASH_LENGTH:
            self.context.logger.error(
                "Something went wrong while trying to get the slash transaction's hash. "
                f"Invalid hash {tx_hash!r} was returned."
            )
            return None

        # strip "0x" from the response hash
        return tx_hash[2:]

    def async_act(self) -> Generator:
        """
        Performs the slashing check logic.

        This method is responsible for checking whether there is any slashable event.
        When an agent detects a slashable event, it prepares a multisend transaction,
        on which the majority of the agents in the service have to reach consensus on.

        :return: None
        :yield: None
        """
        if not self.is_majority_possible() or self.synchronized_data.slashing_in_flight:
            # If the service does not have enough participants to reach consensus, there is no need to run the act.
            # Additionally, we verify whether a slashing operation has already been triggered to avoid duplication.
            yield from self.sleep(self.params.sleep_time)
            return

        if self.params.service_registry_address is None:  # pragma: no cover
            raise ValueError(
                "Service registry address not set, but is required for slashing!"
            )

        self._check_offence_status()
        if len(self._slash_amounts) == 0:
            # no slashable events are present, so we sleep and try again
            yield from self.sleep(self.params.sleep_time)
            return

        self.context.logger.info(
            f"Slashable events detected for agent instances: {self.slashable_instances}. "
            f"Preparing slashing transaction with amounts: {self.slashable_amounts}."
        )

        data = yield from self._get_slash_data()
        if data is None:
            self.context.logger.error(
                "Cannot construct the safe tx without the slash data."
            )
            return

        safe_tx_hash = yield from self._get_safe_tx_hash(data)
        if safe_tx_hash is None:
            self.context.logger.error("Cannot slash without a safe tx hash.")
            return

        slashing_tx_hex = hash_payload_to_hex(
            safe_tx_hash,
            _ETHER_VALUE,
            _SAFE_GAS,
            str(self.params.service_registry_address),
            data,
        )

        slashing_tx_payload = SlashingTxPayload(
            self.context.agent_address, slashing_tx_hex
        )

        self.context.logger.info("Successfully prepared slashing tx payload.")
        yield from self.send_a2a_transaction(slashing_tx_payload)
        yield from self.wait_for_condition(self.is_slashing_majority_reached)


@dataclass
class OperatorSlashedEventLog:
    """The logs' structure of an operator slashed event."""

    amount: int
    operator: str
    serviceId: int


@dataclass
class OperatorSlashedInfo:
    """The slash receipt's information after being parsed."""

    slash_timestamp: int
    events: List[OperatorSlashedEventLog]

    def __post_init__(self) -> None:
        """Post initialization process for the events."""
        if all(isinstance(event, dict) for event in self.events):
            events = cast(dict, self.events)
            self.events = [OperatorSlashedEventLog(**event) for event in events]


class StatusResetBehaviour(SlashingBaseBehaviour):
    """Behaviour that runs after a slash tx has been verified."""

    matching_round = StatusResetRound

    @property
    def slashing_synced_data(self) -> SlashingSyncedData:
        """Return the synchronized data."""
        return cast(SlashingSyncedData, self.shared_state.synchronized_data)

    @property
    def tx_settlement_synced_data(self) -> TxSettlementSyncedData:
        """Return the synchronized data."""
        return TxSettlementSyncedData(db=super().synchronized_data.db)

    def _process_slash_receipt(
        self, slash_tx_hash: str
    ) -> Generator[None, None, Optional[OperatorSlashedInfo]]:
        """Process the slash tx' s receipt."""
        response_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.params.service_registry_address,
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="process_slash_receipt",
            tx_hash=slash_tx_hash,
            chain_id=self.params.default_chain_id,
        )

        if response_msg.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.error(
                "Couldn't process slashing transaction's receipt. Expected response performative "
                f"{ContractApiMessage.Performative.RAW_TRANSACTION.value}, received {response_msg}."  # type: ignore
            )
            return None

        if response_msg.raw_transaction.body is None:  # pragma: no cover
            # error is logged in the contract
            return None

        return OperatorSlashedInfo(**response_msg.raw_transaction.body)

    def _get_instances_mapping(
        self,
        agent_instances: FrozenSet[str],
    ) -> Generator[None, None, Optional[Dict[str, str]]]:
        """
        Retrieve a mapping of the given agent instances to their operators.

        Please keep in mind that this method performs a call for each agent instance.

        :param agent_instances: the agent instances to be mapped.
        :return: a mapping of the given agent instances to their operators.
        """
        # Ideally, `mapOperatorAndServiceIdAgentInstances` should be used instead of `mapAgentInstanceOperators`,
        # so that we have operators mapped to lists of agent instances for the given service id
        response_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.params.service_registry_address,
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="get_operators_mapping",
            agent_instances=agent_instances,
            chain_id=self.params.default_chain_id,
        )

        if response_msg.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.error(
                "Couldn't get operators mapping. Expected response performative "
                f"{ContractApiMessage.Performative.RAW_TRANSACTION.value}, received {response_msg}."  # type: ignore
            )
            return None

        return response_msg.raw_transaction.body

    def async_act(self) -> Generator:
        """
        Performs the slash result check and the status reset logic.

        This behaviour is responsible for checking the `OperatorSlashed` event
        to determine whether the slashing tx was successful.
        If that is the case, it proceeds with resetting the offence status of the slashed agents.
        """
        # the `OperatorSlashed` event returns the operators and not the agent instances, so we will need a mapping
        operators_mapping = self.slashing_synced_data.operators_mapping
        if operators_mapping is None:
            instances_mapping = yield from self._get_instances_mapping(
                self.slashing_synced_data.all_participants
            )
            if instances_mapping is None:
                self.context.logger.error(
                    f"Cannot continue without the agents to operators mapping. "
                    f"Retrying in {self.params.sleep_time} seconds."
                )
                yield from self.sleep(self.params.sleep_time)
                return
            # as also mentioned in `_get_instances_mapping`,
            # it would be better if `mapOperatorAndServiceIdAgentInstances` was used,
            # because getting the inverse here would not be necessary
            # in any case, we need to sort the mapping,
            # as the order in which they are received from the contract is not guaranteed,
            # and we need its content to be deterministic as it is later passed via the payload
            operators_mapping = inverse(dict(sorted(instances_mapping.items())))

        # check `OperatorSlashed` event to see which agents were slashed and if everything went as expected
        slash_info = yield from self._process_slash_receipt(
            self.tx_settlement_synced_data.final_tx_hash
        )
        if slash_info is None:
            # This signifies that we will continue retrying until either we succeed or the round times out.
            # However, this approach carries the risk of getting trapped in a loop!
            #
            # One potential solution could involve introducing a counter to track the number of retries.
            # If the counter would surpass a predefined threshold, we would then proceed to reset the offences
            # providing a mechanism to prevent an indefinite loop.
            self.context.logger.error(
                "Something unexpected happened while checking the slashing operation's result! "
                f"Retrying in {self.params.sleep_time} seconds."
            )
            yield from self.sleep(self.params.sleep_time)
            return

        agent_to_timestamp = {
            agent_instance: slash_info.slash_timestamp
            for event in slash_info.events
            for agent_instance in operators_mapping[event.operator]
        }
        self.context.logger.info(
            f"A slashing operation has been performed for the operator(s) of agents {list(agent_to_timestamp.keys())}."
        )
        self.context.logger.info("Resetting status for the slashed agents.")
        self.offence_status = {
            agent: OffenceStatus() for agent in agent_to_timestamp.keys()
        }
        self.context.logger.info("Successfully reset status for the slashed agents.")

        status_reset_payload = StatusResetPayload(
            self.context.agent_address,
            json.dumps(operators_mapping, sort_keys=True),
            json.dumps(agent_to_timestamp, sort_keys=True),
        )

        yield from self.send_a2a_transaction(status_reset_payload)
        yield from self.wait_until_round_end()
        self.set_done()


class SlashingAbciBehaviours(AbstractRoundBehaviour):
    """This class defines the behaviours required for slashing."""

    initial_behaviour_cls = TransactionSettlementRoundBehaviour.initial_behaviour_cls
    abci_app_cls = SlashingAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        SlashingCheckBehaviour,  # type: ignore
        StatusResetBehaviour,  # type: ignore
        *TransactionSettlementRoundBehaviour.behaviours,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helper functions to manage on-chain services"""

import binascii
import time
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple, cast

from aea.crypto.base import Crypto, LedgerApi
from hexbytes import HexBytes

from autonomy.chain.base import ServiceState, registry_contracts
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.constants import (
    ERC20_CONTRACT,
    GNOSIS_SAFE_PROXY_FACTORY_CONTRACT,
    GNOSIS_SAFE_SAME_ADDRESS_MULTISIG_CONTRACT,
    MULTISEND_CONTRACT,
    SERVICE_MANAGER_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
    SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT,
)
from autonomy.chain.exceptions import (
    ChainInteractionError,
    InstanceRegistrationFailed,
    ServiceDeployFailed,
    ServiceRegistrationFailed,
    TerminateServiceFailed,
    UnbondServiceFailed,
)
from autonomy.chain.mint import transact
from autonomy.chain.tx import TxSettler


NULL_ADDRESS = "0x0000000000000000000000000000000000000000"
DEFAULT_FALLBACK_HANDLER = "0xf48f2b2d2a534e402487b3ee7c18c33aec0fe5e4"
DEFAULT_DEPLOY_PAYLOAD = "0x0000000000000000000000000000000000000000{fallback_handler}000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"

ServiceInfo = Tuple[int, str, bytes, int, int, int, int, List[int]]


class MultiSendOperation(Enum):
    """Operation types."""

    CALL = 0
    DELEGATE_CALL = 1


def get_delployment_payload(fallback_handler: Optional[str] = None) -> str:
    """Calculates deployment payload."""
    return (
        DEFAULT_DEPLOY_PAYLOAD.format(
            fallback_handler=(fallback_handler or DEFAULT_FALLBACK_HANDLER)[2:]
        )
        + int(time.time()).to_bytes(32, "big").hex()
    )


def get_agent_instances(
    ledger_api: LedgerApi, chain_type: ChainType, token_id: int
) -> Dict:
    """
    Get the list of agent instances.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param chain_type: Chain type
    :param token_id: Token ID pointing to the on-chain service
    :returns: number of agent instances and the list of registered addressed
    """

    return registry_contracts.service_registry.get_agent_instances(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.get(SERVICE_REGISTRY_CONTRACT.name).contracts[
            chain_type
        ],
        service_id=token_id,
    )


def get_service_info(
    ledger_api: LedgerApi, chain_type: ChainType, token_id: int
) -> ServiceInfo:
    """
    Returns service info.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param chain_type: Chain type
    :param token_id: Token ID pointing to the on-chain service
    :returns: security deposit, multisig address, IPFS hash for config,
            threshold, max number of agent instances, number of agent instances,
            service state, list of cannonical agents
    """

    return registry_contracts.service_registry.get_service_information(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.get(SERVICE_REGISTRY_CONTRACT.name).contracts[
            chain_type
        ],
        token_id=token_id,
    )


def get_token_deposit_amount(
    ledger_api: LedgerApi,
    chain_type: ChainType,
    service_id: int,
    agent_id: Optional[int] = None,
) -> int:
    """Returns service info."""
    if agent_id is None:
        *_, (agent_id, *_) = get_service_info(
            ledger_api=ledger_api, chain_type=chain_type, token_id=service_id
        )
    return registry_contracts.service_registry_token_utility.get_agent_bond(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.get(
            SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT.name
        ).contracts[chain_type],
        service_id=service_id,
        agent_id=agent_id,
    ).get("bond")


def get_activate_registration_amount(
    ledger_api: LedgerApi,
    chain_type: ChainType,
    service_id: int,
    agents: List[int],
) -> int:
    """Get activate registration amount."""
    agent_to_deposit = {}
    amount = 0
    for agent in agents:
        if agent not in agent_to_deposit:
            agent_to_deposit[agent] = get_token_deposit_amount(
                ledger_api=ledger_api,
                chain_type=chain_type,
                service_id=service_id,
                agent_id=agent,
            )
        amount += agent_to_deposit[agent]
    return amount


def is_service_token_secured(
    ledger_api: LedgerApi,
    chain_type: ChainType,
    service_id: int,
) -> bool:
    """Check if the service is token secured."""
    response = (
        registry_contracts.service_registry_token_utility.is_token_secured_service(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT.name
            ).contracts[chain_type],
            service_id=service_id,
        )
    )
    return response["is_token_secured_service"]


def approve_erc20_usage(  # pylint: disable=too-many-arguments, too-many-locals
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    spender: str,
    amount: int,
    sender: str,
    dry_run: bool = False,
    timeout: Optional[float] = None,
    retries: Optional[int] = None,
    sleep: Optional[float] = None,
) -> None:
    """Approve ERC20 token usage."""
    tx_settler = TxSettler(
        chain_type=chain_type,
        ledger_api=ledger_api,
        crypto=crypto,
        timeout=timeout,
        retries=retries,
        sleep=sleep,
    )
    method = registry_contracts.erc20.get_approve_tx
    contract = ERC20_CONTRACT.name
    kwargs = dict(
        spender=spender,
        amount=amount,
        sender=sender,
    )
    receipt = tx_settler.transact(
        method=method,
        contract=contract,
        kwargs=kwargs,
    )
    if dry_run:  # pragma: nocover
        print("=== Dry run output ===")
        print("Method: " + str(method).split(" ")[2])
        print(f"Contract: {ContractConfigs.get(name=contract).contracts[chain_type]}")
        print("Kwargs: ")
        for key, val in kwargs.items():
            print(f"    {key}: {val}")
        print("Transaction: ")
        for key, val in receipt.items():
            print(f"    {key}: {val}")
        return
    events = cast(
        List[Dict],
        tx_settler.process(
            event="Approval",
            receipt=receipt,
            contract=ERC20_CONTRACT,
        ).get("events"),
    )
    for event in events:
        if event["args"]["spender"] == spender:
            return
    raise ChainInteractionError("Funds approval request failed")


class ServiceManager:
    """Service manager."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        ledger_api: LedgerApi,
        crypto: Crypto,
        chain_type: ChainType,
        dry_run: bool = False,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
    ) -> None:
        """Initialize object."""
        self.ledger_api = ledger_api
        self.crypto = crypto
        self.chain_type = chain_type
        self.timeout = timeout
        self.retries = retries
        self.sleep = sleep
        self.dry_run = dry_run

    def _transact(  # pylint: disable=too-many-arguments
        self,
        method: Callable,
        kwargs: Dict,
        build_tx_ctr: str,
        event: str,
        service_id: int,
        exception: Exception,
    ) -> None:
        """Auxiliary method for managing services on-chain."""
        tx_settler = TxSettler(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=self.chain_type,
            timeout=self.timeout,
            retries=self.retries,
            sleep=self.sleep,
        )
        receipt = tx_settler.transact(
            method=method,
            contract=build_tx_ctr,
            kwargs=kwargs,
            dry_run=self.dry_run,
        )
        if self.dry_run:
            print("=== Dry run output ===")
            print("Method: " + str(method).split(" ")[2])
            print(
                f"Contract: {ContractConfigs.get(name=build_tx_ctr).contracts[self.chain_type]}"
            )
            print("Kwargs: ")
            for key, val in kwargs.items():
                print(f"    {key}: {val}")
            print("Transaction: ")
            for key, val in receipt.items():
                print(f"    {key}: {val}")
            return
        events = cast(
            List[Dict],
            tx_settler.process(
                event=event,
                receipt=receipt,
                contract=SERVICE_REGISTRY_CONTRACT,
            ).get("events"),
        )
        for _event in events:
            if _event["args"]["serviceId"] == service_id:
                return
        raise exception

    def get_service_info(self, token_id: int) -> ServiceInfo:
        """
        Returns service info.

        :param token_id: Token ID pointing to the on-chain service
        :returns: security deposit, multisig address, IPFS hash for config,
                threshold, max number of agent instances, number of agent instances,
                service state, list of cannonical agents
        """

        return get_service_info(
            ledger_api=self.ledger_api,
            chain_type=self.chain_type,
            token_id=token_id,
        )

    def activate(self, service_id: int) -> None:
        """
        Activate service.

        Once you have minted the service on-chain, you'll have to activate the service
        before you can proceed further.

        :param service_id: Service ID retrieved after minting a service
        """

        (
            cost_of_bond,
            *_,
            service_state,
            _,
        ) = self.get_service_info(token_id=service_id)

        if service_state == ServiceState.NON_EXISTENT.value:
            raise ServiceRegistrationFailed("Service does not exist")

        if service_state != ServiceState.PRE_REGISTRATION.value:
            raise ServiceRegistrationFailed("Service must be inactive")

        self._transact(
            method=registry_contracts.service_manager.get_activate_registration_transaction,
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            kwargs=dict(
                owner=self.crypto.address,
                service_id=service_id,
                security_deposit=cost_of_bond,
            ),
            event="ActivateRegistration",
            service_id=service_id,
            exception=ServiceRegistrationFailed("Could not verify activation event"),
        )

    def register_instance(
        self,
        service_id: int,
        instances: List[str],
        agent_ids: List[int],
    ) -> None:
        """
        Register instance.

        Once you have a service with an active registration, you can register agent
        which will be a part of the service deployment. Using this method you can
        register maximum N amounts per agents, N being the number of slots for an agent
        with agent id being `agent_id`.

        Make sure the instance address you provide is not already a part of any service
        and not as same as the service owner.

        :param service_id: Service ID retrieved after minting a service
        :param instances: Address of the agent instance
        :param agent_ids: Agent ID of the agent that you want this instance to be a part
                        of when deployed
        """

        if len(agent_ids) != len(instances):
            raise InstanceRegistrationFailed(
                "Number of agent instances and agent IDs needs to be same"
            )

        (
            cost_of_bond,
            *_,
            service_state,
            _,
        ) = self.get_service_info(token_id=service_id)

        if service_state == ServiceState.NON_EXISTENT.value:
            raise InstanceRegistrationFailed("Service does not exist")

        if service_state != ServiceState.ACTIVE_REGISTRATION.value:
            raise InstanceRegistrationFailed("Service needs to be in active state")

        security_deposit = cost_of_bond * len(agent_ids)
        self._transact(
            method=registry_contracts.service_manager.get_register_instance_transaction,
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            kwargs=dict(
                owner=self.crypto.address,
                service_id=service_id,
                instances=instances,
                agent_ids=agent_ids,
                security_deposit=security_deposit,
            ),
            event="RegisterInstance",
            service_id=service_id,
            exception=InstanceRegistrationFailed("Could not verify registration event"),
        )

    def deploy(
        self,
        service_id: int,
        fallback_handler: Optional[str] = None,
        reuse_multisig: bool = False,
    ) -> None:
        """
        Deploy service.

        Using this method you can deploy a service on-chain once you have activated
        the service and registered the required agent instances.

        :param service_id: Service ID retrieved after minting a service
        :param fallback_handler: Fallback handler address for gnosis safe multisig
        :param reuse_multisig: Use multisig from the previous deployment
        """

        (
            *_,
            service_state,
            _,
        ) = self.get_service_info(token_id=service_id)

        if service_state == ServiceState.NON_EXISTENT.value:
            raise ServiceDeployFailed("Service does not exist")

        if service_state != ServiceState.FINISHED_REGISTRATION.value:
            raise ServiceDeployFailed(
                "Service needs to be in finished registration state"
            )

        if reuse_multisig:
            _deployment_payload, error = get_reuse_multisig_payload(
                ledger_api=self.ledger_api,
                crypto=self.crypto,
                chain_type=self.chain_type,
                service_id=service_id,
            )
            if _deployment_payload is None:
                raise ServiceDeployFailed(error)
            deployment_payload = _deployment_payload
            gnosis_safe_multisig = ContractConfigs.get(
                GNOSIS_SAFE_SAME_ADDRESS_MULTISIG_CONTRACT.name
            ).contracts[self.chain_type]
        else:
            deployment_payload = get_delployment_payload(
                fallback_handler=fallback_handler
            )
            gnosis_safe_multisig = ContractConfigs.get(
                GNOSIS_SAFE_PROXY_FACTORY_CONTRACT.name
            ).contracts[self.chain_type]

        self._transact(
            method=registry_contracts.service_manager.get_service_deploy_transaction,
            kwargs=dict(
                owner=self.crypto.address,
                service_id=service_id,
                gnosis_safe_multisig=gnosis_safe_multisig,
                deployment_payload=deployment_payload,
            ),
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            event="DeployService",
            service_id=service_id,
            exception=ServiceDeployFailed("Could not verify the deploy event."),
        )

    def terminate(self, service_id: int) -> None:
        """
        Terminate service.

        Using this method you can terminate a service on-chain once you have activated
        the service and registered the required agent instances.

        :param service_id: Service ID retrieved after minting a service
        """

        (
            *_,
            service_state,
            _,
        ) = self.get_service_info(token_id=service_id)

        if service_state == ServiceState.NON_EXISTENT.value:
            raise TerminateServiceFailed("Service does not exist")

        if service_state == ServiceState.PRE_REGISTRATION.value:
            raise TerminateServiceFailed("Service not active")

        if service_state == ServiceState.TERMINATED_BONDED.value:
            raise TerminateServiceFailed("Service already terminated")

        self._transact(
            method=registry_contracts.service_manager.get_terminate_service_transaction,
            kwargs=dict(
                owner=self.crypto.address,
                service_id=service_id,
            ),
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            event="TerminateService",
            service_id=service_id,
            exception=TerminateServiceFailed("Could not verify the terminate event."),
        )

    def unbond(self, service_id: int) -> None:
        """
        Unbond service.

        Using this method you can unbond a service on-chain once you have terminated
        the service.

        :param service_id: Service ID retrieved after minting a service
        """

        (
            *_,
            service_state,
            _,
        ) = self.get_service_info(token_id=service_id)

        if service_state == ServiceState.NON_EXISTENT.value:
            raise UnbondServiceFailed("Service does not exist")

        if service_state != ServiceState.TERMINATED_BONDED.value:
            raise UnbondServiceFailed("Service needs to be in terminated-bonded state")

        self._transact(
            method=registry_contracts.service_manager.get_unbond_service_transaction,
            kwargs=dict(
                owner=self.crypto.address,
                service_id=service_id,
            ),
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            event="OperatorUnbond",
            service_id=service_id,
            exception=UnbondServiceFailed("Could not verify the unbond event."),
        )


def get_reuse_multisig_payload(  # pylint: disable=too-many-locals
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
) -> Tuple[Optional[str], Optional[str]]:
    """Reuse multisig."""
    _, multisig_address, _, threshold, *_ = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )
    if multisig_address == NULL_ADDRESS:
        return None, "Cannot reuse multisig, No previous deployment exist!"

    service_owner = crypto.address
    multisend_address = ContractConfigs.get(MULTISEND_CONTRACT.name).contracts[
        chain_type
    ]
    multisig_instance = registry_contracts.gnosis_safe.get_instance(
        ledger_api=ledger_api,
        contract_address=multisig_address,
    )

    # Verify if the service was terminated properly or not
    old_owners = multisig_instance.functions.getOwners().call()
    if len(old_owners) != 1 or service_owner not in old_owners:
        return (
            None,
            "Service was not terminated properly, the service owner should be the only owner of the safe",
        )

    # Build multisend tx to add new instances as owners
    txs = []
    new_owners = cast(
        List[str],
        get_agent_instances(
            ledger_api=ledger_api,
            chain_type=chain_type,
            token_id=service_id,
        ).get("agentInstances"),
    )

    for _owner in new_owners:
        txs.append(
            {
                "to": multisig_address,
                "data": HexBytes(
                    bytes.fromhex(
                        multisig_instance.encodeABI(
                            fn_name="addOwnerWithThreshold",
                            args=[_owner, 1],
                        )[2:]
                    )
                ),
                "operation": MultiSendOperation.CALL,
                "value": 0,
            }
        )

    txs.append(
        {
            "to": multisig_address,
            "data": HexBytes(
                bytes.fromhex(
                    multisig_instance.encodeABI(
                        fn_name="removeOwner",
                        args=[new_owners[0], service_owner, 1],
                    )[2:]
                )
            ),
            "operation": MultiSendOperation.CALL,
            "value": 0,
        }
    )

    txs.append(
        {
            "to": multisig_address,
            "data": HexBytes(
                bytes.fromhex(
                    multisig_instance.encodeABI(
                        fn_name="changeThreshold",
                        args=[threshold],
                    )[2:]
                )
            ),
            "operation": MultiSendOperation.CALL,
            "value": 0,
        }
    )

    multisend_tx = registry_contracts.multisend.get_multisend_tx(
        ledger_api=ledger_api,
        contract_address=multisend_address,
        txs=txs,
    )
    safe_tx_hash = registry_contracts.gnosis_safe.get_raw_safe_transaction_hash(
        ledger_api=ledger_api,
        contract_address=multisig_address,
        to_address=multisend_address,
        value=multisend_tx["value"],
        data=multisend_tx["data"],
        operation=1,
    ).get("tx_hash")
    approve_hash_tx = registry_contracts.gnosis_safe.get_approve_hash_tx(
        ledger_api=ledger_api,
        contract_address=multisig_address,
        tx_hash=safe_tx_hash,
        sender=crypto.address,
    )
    transact(
        ledger_api=ledger_api,
        crypto=crypto,
        tx=approve_hash_tx,
    )

    safe_tx_bytes = binascii.unhexlify(safe_tx_hash[2:])
    owner_to_signature = {
        crypto.address: crypto.sign_message(
            message=safe_tx_bytes,
            is_deprecated_mode=True,
        )[2:]
    }
    signature_bytes = registry_contracts.gnosis_safe.get_packed_signatures(
        owners=tuple(old_owners), signatures_by_owner=owner_to_signature
    )
    safe_exec_data = multisig_instance.encodeABI(
        fn_name="execTransaction",
        args=[
            multisend_address,  # to address
            multisend_tx["value"],  # value
            multisend_tx["data"],  # data
            1,  # operation
            0,  # safe tx gas
            0,  # bas gas
            0,  # safe gas price
            NULL_ADDRESS,  # gas token
            NULL_ADDRESS,  # refund receiver
            signature_bytes,  # signatures
        ],
    )
    payload = multisig_address + safe_exec_data[2:]
    return payload, None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abstract_round_abci' skill."""

from abc import ABC, ABCMeta
from collections import defaultdict
from dataclasses import asdict
from typing import (
    AbstractSet,
    Any,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
)

from aea.skills.base import Behaviour

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbstractRound,
    EventType,
    PendingOffencesPayload,
    PendingOffencesRound,
    PendingOffense,
    RoundSequence,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    TmManager,
    make_degenerate_behaviour,
)
from packages.valory.skills.abstract_round_abci.models import SharedState


SLASHING_BACKGROUND_BEHAVIOUR_ID = "slashing_check_behaviour"
TERMINATION_BACKGROUND_BEHAVIOUR_ID = "background_behaviour"


BehaviourType = Type[BaseBehaviour]
Action = Optional[str]
TransitionFunction = Dict[BehaviourType, Dict[Action, BehaviourType]]


class _MetaRoundBehaviour(ABCMeta):
    """A metaclass that validates AbstractRoundBehaviour's attributes."""

    are_background_behaviours_set: bool = False

    def __new__(mcs, name: str, bases: Tuple, namespace: Dict, **kwargs: Any) -> Type:  # type: ignore
        """Initialize the class."""
        new_cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        if ABC in bases:
            # abstract class, return
            return new_cls
        if not issubclass(new_cls, AbstractRoundBehaviour):
            # the check only applies to AbstractRoundBehaviour subclasses
            return new_cls

        mcs.are_background_behaviours_set = bool(
            new_cls.background_behaviours_cls - {PendingOffencesBehaviour}
        )
        mcs._check_consistency(cast(AbstractRoundBehaviour, new_cls))
        return new_cls

    @classmethod
    def _check_consistency(mcs, behaviour_cls: "AbstractRoundBehaviour") -> None:
        """Check consistency of class attributes."""
        mcs._check_all_required_classattributes_are_set(behaviour_cls)
        mcs._check_behaviour_id_uniqueness(behaviour_cls)
        mcs._check_initial_behaviour_in_set_of_behaviours(behaviour_cls)
        mcs._check_matching_round_consistency(behaviour_cls)

    @classmethod
    def _check_all_required_classattributes_are_set(
        mcs, behaviour_cls: "AbstractRoundBehaviour"
    ) -> None:
        """Check that all the required class attributes are set."""
        try:
            _ = behaviour_cls.abci_app_cls
            _ = behaviour_cls.behaviours
            _ = behaviour_cls.initial_behaviour_cls
        except AttributeError as e:
            raise ABCIAppInternalError(*e.args) from None

    @classmethod
    def _check_behaviour_id_uniqueness(
        mcs, behaviour_cls: "AbstractRoundBehaviour"
    ) -> None:
        """Check that behaviour ids are unique across behaviours."""
        behaviour_id_to_behaviour = defaultdict(lambda: [])
        for behaviour_class in behaviour_cls.behaviours:
            behaviour_id_to_behaviour[behaviour_class.auto_behaviour_id()].append(
                behaviour_class
            )
            if len(behaviour_id_to_behaviour[behaviour_class.auto_behaviour_id()]) > 1:
                behaviour_classes_names = [
                    _behaviour_cls.__name__
                    for _behaviour_cls in behaviour_id_to_behaviour[
                        behaviour_class.auto_behaviour_id()
                    ]
                ]
                raise ABCIAppInternalError(
                    f"behaviours {behaviour_classes_names} have the same behaviour id '{behaviour_class.auto_behaviour_id()}'"
                )

    @classmethod
    def _check_matching_round_consistency(
        mcs, behaviour_cls: "AbstractRoundBehaviour"
    ) -> None:
        """Check that matching rounds are: (1) unique across behaviour, and (2) covering."""
        matching_bg_round_classes = {
            behaviour_cls.matching_round
            for behaviour_cls in behaviour_cls.background_behaviours_cls
        }
        round_to_behaviour: Dict[Type[AbstractRound], List[BehaviourType]] = {
            round_cls: []
            for round_cls in behaviour_cls.abci_app_cls.get_all_round_classes(
                matching_bg_round_classes,
                mcs.are_background_behaviours_set,
            )
        }

        # check uniqueness
        for b in behaviour_cls.behaviours:
            behaviours = round_to_behaviour.get(b.matching_round, None)
            if behaviours is None:
                raise ABCIAppInternalError(
                    f"Behaviour {b.behaviour_id!r} specifies unknown {b.matching_round!r} as a matching round. "
                    "Please make sure that the round is implemented and belongs to the FSM. "
                    f"If {b.behaviour_id!r} is a background behaviour, please make sure that it is set correctly, "
                    f"by overriding the corresponding attribute of the chained skill's behaviour."
                )
            behaviours.append(b)
            if len(behaviours) > 1:
                behaviour_cls_ids = [
                    behaviour_cls_.auto_behaviour_id() for behaviour_cls_ in behaviours
                ]
                raise ABCIAppInternalError(
                    f"behaviours {behaviour_cls_ids} have the same matching round '{b.matching_round.auto_round_id()}'"
                )

        # check covering
        for round_cls, behaviours in round_to_behaviour.items():
            if round_cls in behaviour_cls.abci_app_cls.final_states:
                if len(behaviours) != 0:
                    raise ABCIAppInternalError(
                        f"round {round_cls.auto_round_id()} is a final round it shouldn't have any matching behaviours."
                    )
            elif len(behaviours) == 0:
                raise ABCIAppInternalError(
                    f"round {round_cls.auto_round_id()} is not a matching round of any behaviour"
                )

    @classmethod
    def _check_initial_behaviour_in_set_of_behaviours(
        mcs, behaviour_cls: "AbstractRoundBehaviour"
    ) -> None:
        """Check the initial behaviour is in the set of behaviours."""
        if behaviour_cls.initial_behaviour_cls not in behaviour_cls.behaviours:
            raise ABCIAppInternalError(
                f"initial behaviour {behaviour_cls.initial_behaviour_cls.auto_behaviour_id()} is not in the set of behaviours"
            )


class PendingOffencesBehaviour(BaseBehaviour):
    """A behaviour responsible for checking whether there are any pending offences."""

    matching_round = PendingOffencesRound

    @property
    def round_sequence(self) -> RoundSequence:
        """Get the round sequence from the shared state."""
        return cast(SharedState, self.context.state).round_sequence

    @property
    def pending_offences(self) -> Set[PendingOffense]:
        """Get the pending offences from the round sequence."""
        return self.round_sequence.pending_offences

    def has_pending_offences(self) -> bool:
        """Check if there are any pending offences."""
        return bool(len(self.pending_offences))

    def async_act(self) -> Generator:
        """
        Checks the pending offences.

        This behaviour simply checks if the set of pending offences is not empty.
        When it’s not empty, it pops the offence from the set, and sends it to the rest of the agents via a payload

        :return: None
        :yield: None
        """
        yield from self.wait_for_condition(self.has_pending_offences)
        offence = self.pending_offences.pop()
        offence_detected_log = (
            f"An offence of type {offence.offense_type.name} has been detected "
            f"for agent with address {offence.accused_agent_address} during round {offence.round_count}. "
        )
        offence_expiration = offence.last_transition_timestamp + offence.time_to_live
        last_timestamp = self.round_sequence.last_round_transition_timestamp

        if offence_expiration < last_timestamp.timestamp():
            ignored_log = "Offence will be ignored as it has expired."
            self.context.logger.info(offence_detected_log + ignored_log)
            return

        sharing_log = "Sharing offence with the other agents."
        self.context.logger.info(offence_detected_log + sharing_log)

        payload = PendingOffencesPayload(
            self.context.agent_address, *asdict(offence).values()
        )
        yield from self.send_a2a_transaction(payload)
        yield from self.wait_until_round_end()
        self.set_done()


class AbstractRoundBehaviour(  # pylint: disable=too-many-instance-attributes
    Behaviour, ABC, Generic[EventType], metaclass=_MetaRoundBehaviour
):
    """This behaviour implements an abstract round behaviour."""

    abci_app_cls: Type[AbciApp[EventType]]
    behaviours: AbstractSet[BehaviourType]
    initial_behaviour_cls: BehaviourType
    background_behaviours_cls: Set[BehaviourType] = {PendingOffencesBehaviour}  # type: ignore

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the behaviour."""
        super().__init__(**kwargs)
        self._behaviour_id_to_behaviours: Dict[
            str, BehaviourType
        ] = self._get_behaviour_id_to_behaviour_mapping(self.behaviours)
        self._round_to_behaviour: Dict[
            Type[AbstractRound], BehaviourType
        ] = self._get_round_to_behaviour_mapping(self.behaviours)

        self.current_behaviour: Optional[BaseBehaviour] = None
        self.background_behaviours: Set[BaseBehaviour] = set()
        self.tm_manager: Optional[TmManager] = None
        # keep track of last round height so to detect changes
        self._last_round_height = 0

        # this variable remembers the actual next transition
        # when we cannot preemptively interrupt the current behaviour
        # because it has not a matching round.
        self._next_behaviour_cls: Optional[BehaviourType] = None

    @classmethod
    def _get_behaviour_id_to_behaviour_mapping(
        cls, behaviours: AbstractSet[BehaviourType]
    ) -> Dict[str, BehaviourType]:
        """Get behaviour id to behaviour mapping."""
        result: Dict[str, BehaviourType] = {}
        for behaviour_cls in behaviours:
            behaviour_id = behaviour_cls.auto_behaviour_id()
            if behaviour_id in result:
                raise ValueError(
                    f"cannot have two behaviours with the same id; got {behaviour_cls} and {result[behaviour_id]} both with id '{behaviour_id}'"
                )
            result[behaviour_id] = behaviour_cls
        return result

    @classmethod
    def _get_round_to_behaviour_mapping(
        cls, behaviours: AbstractSet[BehaviourType]
    ) -> Dict[Type[AbstractRound], BehaviourType]:
        """Get round-to-behaviour mapping."""
        result: Dict[Type[AbstractRound], BehaviourType] = {}
        for behaviour_cls in behaviours:
            round_cls = behaviour_cls.matching_round
            if round_cls in result:
                raise ValueError(
                    f"the behaviours '{behaviour_cls.auto_behaviour_id()}' and '{result[round_cls].auto_behaviour_id()}' point to the same matching round '{round_cls.auto_round_id()}'"
                )
            result[round_cls] = behaviour_cls

        # iterate over rounds and map final (i.e. degenerate) rounds
        #  to the degenerate behaviour class
        for final_round_cls in cls.abci_app_cls.final_states:
            new_degenerate_behaviour = make_degenerate_behaviour(final_round_cls)
            new_degenerate_behaviour.matching_round = final_round_cls
            result[final_round_cls] = new_degenerate_behaviour

        return result

    def instantiate_behaviour_cls(self, behaviour_cls: BehaviourType) -> BaseBehaviour:
        """Instantiate the behaviours class."""
        return behaviour_cls(
            name=behaviour_cls.auto_behaviour_id(), skill_context=self.context
        )

    def _setup_background(self) -> None:
        """Set up the background behaviours."""
        params = cast(BaseBehaviour, self.current_behaviour).params
        for background_cls in self.background_behaviours_cls:
            background_cls = cast(Type[BaseBehaviour], background_cls)

            if (
                not params.use_termination
                and background_cls.auto_behaviour_id()
                == TERMINATION_BACKGROUND_BEHAVIOUR_ID
            ) or (
                not params.use_slashing
                and background_cls.auto_behaviour_id()
                == SLASHING_BACKGROUND_BEHAVIOUR_ID
                or background_cls == PendingOffencesBehaviour
            ):
                # comparing with the behaviour id is not entirely safe, as there is a potential for conflicts
                # if a user creates a behaviour with the same name
                continue

            self.background_behaviours.add(
                self.instantiate_behaviour_cls(background_cls)
            )

    def setup(self) -> None:
        """Set up the behaviours."""
        self.current_behaviour = self.instantiate_behaviour_cls(
            self.initial_behaviour_cls
        )
        self.tm_manager = self.instantiate_behaviour_cls(TmManager)  # type: ignore
        self._setup_background()

    def teardown(self) -> None:
        """Tear down the behaviour"""

    def _background_act(self) -> None:
        """Call the act wrapper for the background behaviours."""
        for behaviour in self.background_behaviours:
            behaviour.act_wrapper()

    def act(self) -> None:
        """Implement the behaviour."""
        tm_manager = cast(TmManager, self.tm_manager)
        if tm_manager.tm_communication_unhealthy or tm_manager.is_acting:
            # tendermint is not healthy, or we are already applying a fix.
            # try_fix() internally uses generators, that's why it's relevant
            # to know whether a fix is already being applied.
            # It might happen that tendermint is healthy, but the fix is not yet finished.
            tm_manager.try_fix()
            return

        tm_manager.informed = False
        tm_manager.acn_communication_attempted = False
        self._process_current_round()
        if self.current_behaviour is None:
            return

        self.current_behaviour.act_wrapper()
        if self.current_behaviour.is_done():
            self.current_behaviour.clean_up()
            self.current_behaviour = None

        self._background_act()

    def _process_current_round(self) -> None:
        """Process current ABCIApp round."""
        current_round_height = self.context.state.round_sequence.current_round_height
        if (
            self.current_behaviour is not None
            and self._last_round_height == current_round_height
        ):
            # round has not changed - do nothing
            return
        self._last_round_height = current_round_height
        current_round_cls = type(self.context.state.round_sequence.current_round)

        # each round has a behaviour associated to it
        next_behaviour_cls = self._round_to_behaviour[current_round_cls]

        # stop the current behaviour and replace it with the new behaviour
        if self.current_behaviour is not None:
            current_behaviour = cast(BaseBehaviour, self.current_behaviour)
            current_behaviour.clean_up()
            current_behaviour.stop()
            self.context.logger.debug(
                "overriding transition: current behaviour: '%s', next behaviour: '%s'",
                self.current_behaviour.behaviour_id if self.current_behaviour else None,
                next_behaviour_cls.auto_behaviour_id(),
            )

        self.current_behaviour = self.instantiate_behaviour_cls(next_behaviour_cls)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base deployments module."""
import abc
import json
import logging
import os
from base64 import urlsafe_b64encode
from copy import copy, deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast
from warnings import warn

from aea.cli.generate_key import _generate_multiple_keys
from aea.configurations.base import (
    ConnectionConfig,
    ContractConfig,
    ProtocolConfig,
    SkillConfig,
)
from aea.configurations.constants import ADDRESS, LEDGER, PRIVATE_KEY, SKILL
from aea.configurations.data_types import PackageType, PublicId
from aea.helpers.env_vars import apply_env_variables
from typing_extensions import TypedDict

from autonomy.analyse.service import ABCI
from autonomy.configurations.base import Service
from autonomy.configurations.loader import load_service_config
from autonomy.constants import DEFAULT_DOCKER_IMAGE_AUTHOR
from autonomy.deploy.constants import DEFAULT_ENCODING, INFO


ENV_VAR_ID = "ID"
ENV_VAR_AEA_AGENT = "AEA_AGENT"
ENV_VAR_LOG_LEVEL = "LOG_LEVEL"
ENV_VAR_AEA_PASSWORD = "AEA_PASSWORD"  # nosec
ENV_VAR_DEPENDENCIES = "DEPENDENCIES"  # nosec
ENV_VAR_OPEN_AUTONOMY_TM_WRITE_TO_LOG = "OPEN_AUTONOMY_TM_WRITE_TO_LOG"

PARAM_ARGS_PATH = ("models", "params", "args")
SETUP_PARAM_PATH = (*PARAM_ARGS_PATH, "setup")
SAFE_CONTRACT_ADDRESS = "safe_contract_address"
ALL_PARTICIPANTS = "all_participants"
CONSENSUS_THRESHOLD = "consensus_threshold"


DEFAULT_ABCI_PORT = 26658


KUBERNETES_DEPLOYMENT = "kubernetes"
DOCKER_COMPOSE_DEPLOYMENT = "docker-compose"
LOCALHOST_DEPLOYMENT = "localhost"

LOOPBACK = "127.0.0.1"
LOCALHOST = "localhost"
TENDERMINT_P2P_PORT = 26656

TENDERMINT_NODE = "http://{host}:26657"
TENDERMINT_COM = "http://{host}:8080"

TENDERMINT_NODE_LOCAL = f"http://{LOCALHOST}:26657"
TENDERMINT_COM_LOCAL = f"http://{LOCALHOST}:8080"

TENDERMINT_URL_PARAM = "tendermint_url"
TENDERMINT_COM_URL_PARAM = "tendermint_com_url"

TENDERMINT_P2P_URL = "{host}:{port}"
TENDERMINT_P2P_URL_PARAM = "tendermint_p2p_url"
TENDERMINT_P2P_URL_ENV_VAR = "TM_P2P_NODE_URL_{}"

COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}

DEFAULT_AGENT_MEMORY_LIMIT = int(os.environ.get("AUTONOMY_AGENT_MEMORY_LIMIT", 1024))
DEFAULT_AGENT_CPU_LIMIT = float(os.environ.get("AUTONOMY_AGENT_CPU_LIMIT", 1.0))
DEFAULT_AGENT_MEMORY_REQUEST = int(os.environ.get("AUTONOMY_AGENT_MEMORY_REQUEST", 256))
DEFAULT_AGENT_CPU_REQUEST = float(os.environ.get("AUTONOMY_AGENT_CPU_REQUEST", 1.0))


def tm_write_to_log() -> bool:
    """Check the environment variable to see if the user wants to write to log file or not."""
    return os.getenv(ENV_VAR_OPEN_AUTONOMY_TM_WRITE_TO_LOG, "true").lower() == "true"


class NotValidKeysFile(Exception):
    """Raise when provided keys file is not valid."""


class ResourceValues(TypedDict):
    """Resource type."""

    cpu: Optional[float]
    memory: Optional[int]


class Resource(TypedDict):
    """Resource values."""

    requested: ResourceValues
    limit: ResourceValues


class Resources(TypedDict):
    """Deployment resources."""

    agent: Resource


DEFAULT_RESOURCE_VALUES = Resources(
    {
        "agent": {
            "limit": {
                "cpu": DEFAULT_AGENT_CPU_LIMIT,
                "memory": DEFAULT_AGENT_MEMORY_LIMIT,
            },
            "requested": {
                "cpu": DEFAULT_AGENT_CPU_REQUEST,
                "memory": DEFAULT_AGENT_MEMORY_REQUEST,
            },
        }
    }
)


def build_hash_id() -> str:
    """Generate a random 4 character hash id for the deployment build directory name."""
    return urlsafe_b64encode(bytes(os.urandom(3))).decode()


class ServiceBuilder:  # pylint: disable=too-many-instance-attributes
    """Class to assist with generating deployments."""

    deplopyment_type: str = DOCKER_COMPOSE_DEPLOYMENT
    log_level: str = INFO

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service: Service,
        keys: Optional[List[Union[List[Dict[str, str]], Dict[str, str]]]] = None,
        agent_instances: Optional[List[str]] = None,
        apply_environment_variables: bool = False,
        service_hash_id: Optional[str] = None,
        service_offset: int = 0,
    ) -> None:
        """Initialize the Base Deployment."""
        if apply_environment_variables:
            warn(  # pragma: no cover
                "`apply_environment_variables` argument is deprecated and will be removed in v1.0.0, "
                "usage of environment variables is default now.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.service = service
        self.service_hash_id = (
            build_hash_id() if service_hash_id is None else service_hash_id
        )
        self.service_offset = service_offset

        self._service_name_clean = (
            self.service.name.replace("_", "") + self.service_hash_id
        )
        self._keys = keys or []
        self._agent_instances = agent_instances
        self._all_participants = self.try_get_all_participants()
        self.multiledger = False

    def get_abci_container_name(self, index: int) -> str:
        """Format ABCI container name."""
        return f"{self._service_name_clean}_abci_{index}"

    def get_tm_container_name(self, index: int) -> str:
        """Format tendermint container name."""
        return f"{self._service_name_clean}_tm_{index}"

    def try_get_all_participants(self) -> Optional[List[str]]:
        """Try get all participants from the ABCI overrides"""
        for override in deepcopy(self.service.overrides):
            try:
                (
                    override,
                    component_id,
                    has_multiple_overrides,
                ) = self.service.process_metadata(
                    configuration=override,
                )
                if (
                    component_id.component_type.value == SKILL
                    and has_multiple_overrides
                ):
                    override, *_ = override.values()

                override = apply_env_variables(
                    data=override,
                    env_variables=os.environ.copy(),
                )
                setup_param = self._get_config_from_json_path(
                    override_dict=override, json_path=SETUP_PARAM_PATH
                )
                all_participants = setup_param.get(ALL_PARTICIPANTS)
                if all_participants is not None:
                    return all_participants
            except KeyError:
                continue
        return None

    @property
    def agent_instances(
        self,
    ) -> Optional[List[str]]:
        """Agent instances."""

        return self._agent_instances

    @agent_instances.setter
    def agent_instances(self, instances: List[str]) -> None:
        """Agent instances setter."""

        if self.keys:
            self.verify_agent_instances(
                addresses=set(
                    self._get_addresses(
                        keys=self.keys,
                        multiledger=self.multiledger,
                    )
                ),
                agent_instances=instances,
            )

        self._agent_instances = instances

    @property
    def keys(
        self,
    ) -> List[Union[List[Dict[str, str]], Dict[str, str]]]:
        """Keys."""
        return self._keys

    @classmethod
    def from_dir(  # pylint: disable=too-many-arguments
        cls,
        path: Path,
        keys_file: Optional[Path] = None,
        number_of_agents: Optional[int] = None,
        agent_instances: Optional[List[str]] = None,
        apply_environment_variables: bool = False,
        dev_mode: bool = False,
        service_hash_id: Optional[str] = None,
        service_offset: int = 0,
    ) -> "ServiceBuilder":
        """Service builder from path."""
        service = load_service_config(
            service_path=path,
        )
        if number_of_agents is not None:
            service.number_of_agents = number_of_agents

        service_builder = cls(
            service=service,
            service_hash_id=service_hash_id,
            service_offset=service_offset,
            apply_environment_variables=apply_environment_variables,
        )

        if keys_file is not None:
            if dev_mode and not keys_file.exists():
                # TODO: Add multiledger support
                _generate_multiple_keys(
                    n=service.number_of_agents,
                    type_="ethereum",
                    file=keys_file,
                )
            service_builder.read_keys(keys_file=keys_file)

        if agent_instances is not None:
            service_builder.agent_instances = agent_instances

        return service_builder

    @staticmethod
    def _get_addresses(
        keys: List[Union[List[Dict[str, str]], Dict[str, str]]],
        multiledger: bool = False,
    ) -> List[str]:
        """Get ethereum addresses"""
        if multiledger:
            addresses = []
            for i, _keys in enumerate(cast(List[List[Dict[str, str]]], keys)):
                for _keypair in _keys:
                    if _keypair["ledger"] == "ethereum":
                        addresses.append(_keypair["address"])
                if len(addresses) != i + 1:
                    raise ValueError(f"Ethereum key not found in keyset: {_keys}")
            return addresses
        return [kp["address"] for kp in cast(List[Dict[str, str]], keys)]

    @staticmethod
    def verify_agent_instances(addresses: Set[str], agent_instances: List[str]) -> None:
        """Cross verify agent instances with the keys."""
        instances = set(agent_instances)
        key_not_in_instances = addresses.difference(instances)
        if key_not_in_instances:
            raise NotValidKeysFile(
                f"Key file contains keys which are not registered as instances; invalid keys={key_not_in_instances}"
            )

        instances_not_in_keys = instances.difference(addresses)
        if instances_not_in_keys:
            logging.warning(
                f"Key file does not contain key pair for following instances {instances_not_in_keys}"
            )

        keys_found_with_instances = instances.intersection(addresses)
        logging.info(
            f"Found following keys with registered instances {keys_found_with_instances}"
        )

    @staticmethod
    def _validate_keypair(keypair: Dict, ledger_required: bool = False) -> None:
        """Validate keys set."""
        if ledger_required:
            if {ADDRESS, PRIVATE_KEY, LEDGER} != set(keypair.keys()):
                raise NotValidKeysFile("Key file incorrectly formatted")
            return

        if {ADDRESS, PRIVATE_KEY} != set(keypair.keys()) and {
            ADDRESS,
            PRIVATE_KEY,
            LEDGER,
        } != set(keypair.keys()):
            raise NotValidKeysFile("Key file incorrectly formatted.")

    def read_keys(self, keys_file: Path) -> None:
        """Read in keys from a file on disk."""

        try:
            keys = json.loads(keys_file.read_text(encoding=DEFAULT_ENCODING))
        except json.decoder.JSONDecodeError as e:  # pragma: nocover
            raise NotValidKeysFile(
                "Error decoding keys file, please check the content of the file"
            ) from e

        if isinstance(keys[0], list):
            self.multiledger = True
            for _keys in keys:
                for _keypair in _keys:
                    self._validate_keypair(keypair=_keypair, ledger_required=True)
        else:
            for _keypair in keys:
                self._validate_keypair(keypair=_keypair)

        if self.agent_instances is not None:
            self.verify_agent_instances(
                addresses=set(
                    self._get_addresses(
                        keys=keys,
                        multiledger=self.multiledger,
                    )
                ),
                agent_instances=self.agent_instances,
            )
            self.service.number_of_agents = len(keys)

        if self._all_participants is not None and len(self._all_participants) > 0:
            unwanted_keys = set(
                self._get_addresses(
                    keys=keys,
                    multiledger=self.multiledger,
                )
            ) - set(self._all_participants)
            if len(unwanted_keys) > 0:
                raise NotValidKeysFile(
                    f"Key file contains keys which are not a part of the `all_participants` parameter; keys={unwanted_keys}"
                )
            self.service.number_of_agents = len(keys)

        if self.service.number_of_agents > len(keys):
            raise NotValidKeysFile(
                "Number of agents cannot be greater than available keys."
            )

        self._keys = keys

    @staticmethod
    def _get_config_from_json_path(
        override_dict: Dict,
        json_path: Tuple[str, ...],
    ) -> Dict:
        """Returns `setup` param dict"""

        _key_0, *_keys = json_path
        config = override_dict[_key_0]
        for key in _keys:
            config = config[key]
        return config

    @classmethod
    def _try_update_setup_data(
        cls,
        data: List[Tuple[str, Any]],
        override: Dict,
        skill_id: PublicId,
        has_multiple_overrides: bool = False,
    ) -> None:
        """Try update the setup parameters"""

        def _try_update_setup_params(
            setup_param: Optional[Dict],
            setup_data: List[Tuple[str, Any]],
        ) -> None:
            """Update"""
            if setup_param is None:  # pragma: nocover
                return

            for param, value in setup_data:
                setup_param[param] = value

        try:
            if not has_multiple_overrides:
                setup_param = cls._get_config_from_json_path(
                    override_dict=override, json_path=SETUP_PARAM_PATH
                )
                _try_update_setup_params(setup_param=setup_param, setup_data=data)
                return

            for agent_idx in override:
                setup_param = cls._get_config_from_json_path(
                    override_dict=override[agent_idx], json_path=SETUP_PARAM_PATH
                )
                _try_update_setup_params(setup_param=setup_param, setup_data=data)
        except KeyError:
            logging.warning(
                f"Could not update the setup parameter for {skill_id}; "
                f"Configuration does not contain the json path to the setup parameter"
            )

    def _try_update_tendermint_params(
        self,
        override: Dict,
        skill_id: PublicId,
        has_multiple_overrides: bool = False,
    ) -> None:
        """Try update the tendermint parameters"""

        is_kubernetes_deployment = self.deplopyment_type == KUBERNETES_DEPLOYMENT
        is_localhost_deployment = self.deplopyment_type == LOCALHOST_DEPLOYMENT

        def _update_tendermint_params(
            param_args: Dict,
            idx: int,
        ) -> None:
            """Update tendermint params"""
            if is_kubernetes_deployment or is_localhost_deployment:
                param_args[TENDERMINT_URL_PARAM] = TENDERMINT_NODE_LOCAL
                param_args[TENDERMINT_COM_URL_PARAM] = TENDERMINT_COM_LOCAL
            else:
                param_args[TENDERMINT_URL_PARAM] = TENDERMINT_NODE.format(
                    host=self.get_tm_container_name(index=idx)
                )
                param_args[TENDERMINT_COM_URL_PARAM] = TENDERMINT_COM.format(
                    host=self.get_tm_container_name(index=idx)
                )

            if TENDERMINT_P2P_URL_PARAM not in param_args:
                tm_p2p_url = os.environ.get(
                    TENDERMINT_P2P_URL_ENV_VAR.format(idx),
                    TENDERMINT_P2P_URL.format(
                        host=self.get_tm_container_name(index=idx),
                        port=TENDERMINT_P2P_PORT,
                    ),
                )
                param_args[TENDERMINT_P2P_URL_PARAM] = tm_p2p_url

        try:
            if self.service.number_of_agents == 1:
                if has_multiple_overrides:
                    for agent_idx in override:
                        param_args = self._get_config_from_json_path(
                            override_dict=override[agent_idx], json_path=PARAM_ARGS_PATH
                        )
                        _update_tendermint_params(
                            param_args=param_args,
                            idx=0,
                        )
                else:
                    param_args = self._get_config_from_json_path(
                        override_dict=override, json_path=PARAM_ARGS_PATH
                    )
                    _update_tendermint_params(
                        param_args=param_args,
                        idx=0,
                    )
                return

            if not has_multiple_overrides:
                _base_overrride = deepcopy(override)
                override.clear()
                for i in range(self.get_maximum_participants()):
                    override[i] = deepcopy(_base_overrride)

            for agent_idx in override:
                param_args = self._get_config_from_json_path(
                    override_dict=override[agent_idx], json_path=PARAM_ARGS_PATH
                )
                _update_tendermint_params(
                    param_args=param_args,
                    idx=agent_idx,
                )
        except KeyError:  # pragma: nocover
            logging.warning(
                f"Could not update the tendermint parameter for {skill_id}; "
                f"Configuration does not contain the json path to the setup parameter"
            )

    def try_update_runtime_params(
        self,
        multisig_address: Optional[str] = None,
        agent_instances: Optional[List[str]] = None,
        consensus_threshold: Optional[int] = None,
    ) -> None:
        """Try and update setup parameters."""

        param_overrides: List[Tuple[str, Any]] = []
        if multisig_address is not None:
            param_overrides.append(
                (SAFE_CONTRACT_ADDRESS, multisig_address),
            )

        if agent_instances is not None:
            param_overrides.append(
                (ALL_PARTICIPANTS, agent_instances),
            )

        if consensus_threshold is not None:
            param_overrides.append(
                (CONSENSUS_THRESHOLD, consensus_threshold),
            )

        overrides = copy(self.service.overrides)
        for override in overrides:
            (
                override,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration=override,
            )

            if component_id.component_type.value == SKILL:
                self._try_update_setup_data(
                    data=param_overrides,
                    override=override,
                    skill_id=component_id.public_id,
                    has_multiple_overrides=has_multiple_overrides,
                )
                self._try_update_tendermint_params(
                    override=override,
                    skill_id=component_id.public_id,
                    has_multiple_overrides=has_multiple_overrides,
                )

            override["type"] = component_id.package_type.value
            override["public_id"] = str(component_id.public_id)

        self.service.overrides = overrides

    def get_maximum_participants(self) -> int:
        """Returns the maximum number of participants"""

        if self._all_participants is not None and len(self._all_participants) > 0:
            return len(self._all_participants)

        if self.agent_instances is not None:
            return len(self.agent_instances)

        return max(len(self.keys), self.service.number_of_agents)

    def _update_abci_connection_config(
        self,
        overrides: Dict,
        has_multiple_overrides: bool = False,
    ) -> Dict:
        """Update ABCI connection config"""

        processed_overrides = deepcopy(overrides)
        if self.service.number_of_agents == 1:
            processed_overrides["config"]["host"] = (
                self.get_abci_container_name(index=0)
                if self.deplopyment_type == DOCKER_COMPOSE_DEPLOYMENT
                else LOOPBACK
            )
            processed_overrides["config"]["port"] = processed_overrides["config"].get(
                "port", DEFAULT_ABCI_PORT
            )
            return processed_overrides

        if not has_multiple_overrides:
            processed_overrides = {}
            for i in range(self.get_maximum_participants()):
                processed_overrides[i] = deepcopy(overrides)

        for idx, override in processed_overrides.items():
            override["config"]["host"] = (
                self.get_abci_container_name(index=idx)
                if self.deplopyment_type == DOCKER_COMPOSE_DEPLOYMENT
                else LOOPBACK
            )
            override["config"]["port"] = override["config"].get(
                "port", DEFAULT_ABCI_PORT
            )
            processed_overrides[idx] = override

        return processed_overrides

    def try_update_abci_connection_params(
        self,
    ) -> None:
        """Try and update ledger connection parameters."""

        for override in deepcopy(self.service.overrides):
            (
                override,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration=override,
            )

            if (
                component_id.package_type == PackageType.CONNECTION
                and component_id.name == ABCI
            ):
                service_has_connection_overrides = True
                abci_connection_overrides = override
                break
        else:
            service_has_connection_overrides = False
            (
                abci_connection_overrides,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration={
                    "public_id": "valory/abci:0.1.0",
                    "type": PackageType.CONNECTION.value,
                    "config": {
                        "host": LOCALHOST,
                        "port": DEFAULT_ABCI_PORT,
                    },
                }
            )
        processed_overrides = self._update_abci_connection_config(
            overrides=abci_connection_overrides,
            has_multiple_overrides=has_multiple_overrides,
        )
        processed_overrides["public_id"] = str(component_id.public_id)
        processed_overrides["type"] = PackageType.CONNECTION.value
        service_overrides = deepcopy(self.service.overrides)
        if service_has_connection_overrides:
            service_overrides = [
                override
                for override in service_overrides
                if override["public_id"] != str(component_id.public_id)
                or override["type"] != PackageType.CONNECTION.value
            ]

        service_overrides.append(processed_overrides)
        self.service.overrides = service_overrides

    def process_component_overrides(self, agent_n: int) -> Dict:
        """Generates env vars based on model overrides."""
        final_overrides = {}
        for component_configuration_json in self.service.overrides:
            env_var_dict = self.service.process_component_overrides(
                agent_n, component_configuration_json
            )
            final_overrides.update(env_var_dict)
        return final_overrides

    def generate_agents(self) -> List:
        """Generate multiple agent."""
        addresses = self._get_addresses(
            keys=self.keys,
            multiledger=self.multiledger,
        )
        if self._all_participants is not None and len(self._all_participants) > 0:
            idx_mappings = {
                address: i for i, address in enumerate(self._all_participants)
            }
            agent_override_idx = [
                (i, idx_mappings[kp]) for i, kp in enumerate(addresses)
            ]
            return [self.generate_agent(i, idx) for i, idx in agent_override_idx]
        if self.agent_instances is None:
            return [
                self.generate_agent(i) for i in range(self.service.number_of_agents)
            ]
        idx_mappings = {address: i for i, address in enumerate(self.agent_instances)}
        agent_override_idx = [(i, idx_mappings[kp]) for i, kp in enumerate(addresses)]
        return [self.generate_agent(i, idx) for i, idx in agent_override_idx]

    def generate_common_vars(self, agent_n: int) -> Dict:
        """Retrieve vars common for agent."""
        agent_vars = {
            ENV_VAR_ID: agent_n,
            ENV_VAR_AEA_AGENT: self.service.agent,
            ENV_VAR_LOG_LEVEL: self.log_level,
            ENV_VAR_AEA_PASSWORD: os.environ.get(
                "OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD", ""
            ),
        }
        return agent_vars

    def generate_agent(
        self, agent_n: int, override_idx: Optional[int] = None
    ) -> Dict[Any, Any]:
        """Generate next agent."""
        agent_vars = self.generate_common_vars(agent_n)
        if len(self.service.overrides) == 0:
            return agent_vars

        if override_idx is None:
            override_idx = agent_n

        agent_vars.update(
            self.process_component_overrides(agent_n=override_idx),
        )
        return agent_vars


class BaseDeploymentGenerator(abc.ABC):  # pylint: disable=too-many-instance-attributes
    """Base Deployment Class."""

    service_builder: ServiceBuilder
    output_name: str
    deployment_type: str
    build_dir: Path
    output: str
    tendermint_job_config: Optional[str]
    dev_mode: bool
    use_tm_testnet_setup: bool

    packages_dir: Optional[Path]
    open_aea_dir: Optional[Path]
    resources: Resources

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service_builder: ServiceBuilder,
        build_dir: Path,
        use_tm_testnet_setup: bool = False,
        dev_mode: bool = False,
        packages_dir: Optional[Path] = None,
        open_aea_dir: Optional[Path] = None,
        image_author: Optional[str] = None,
        resources: Optional[Resources] = None,
    ):
        """Initialise with only kwargs."""

        self.service_builder = service_builder
        self.build_dir = build_dir
        self.use_tm_testnet_setup = use_tm_testnet_setup
        self.dev_mode = dev_mode
        self.packages_dir = packages_dir
        self.open_aea_dir = open_aea_dir

        self.tendermint_job_config: Optional[str] = None
        self.image_author = image_author or DEFAULT_DOCKER_IMAGE_AUTHOR
        self.resources = resources if resources is not None else DEFAULT_RESOURCE_VALUES

    @abc.abstractmethod
    def generate(
        self,
        image_version: Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def generate_config_tendermint(self) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def populate_private_keys(
        self,
    ) -> "BaseDeploymentGenerator":
        """Populate the private keys to the deployment."""

    def write_config(self) -> "BaseDeploymentGenerator":
        """Write output to build dir"""

        with open(self.build_dir / self.output_name, "w", encoding="utf8") as f:
            f.write(self.output)
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test tx settlement tool."""

from typing import Any
from unittest import mock

import pytest
from requests.exceptions import ConnectionError as RequestsConnectionError

from autonomy.chain.config import ChainType
from autonomy.chain.exceptions import ChainInteractionError, ChainTimeoutError, RPCError
from autonomy.chain.tx import TxSettler


def test_rpc_error() -> None:
    """Test RPC error."""
    settler = TxSettler(
        ledger_api=mock.Mock(),
        crypto=mock.Mock(),
        chain_type=ChainType.LOCAL,
    )

    def _waitable(*args: Any, **kwargs: Any) -> Any:
        raise RequestsConnectionError()

    with pytest.raises(RPCError):
        settler.transact(_waitable, "service_registry", {})


def test_none_retriable_exception() -> None:
    """Test none retriable exception."""
    settler = TxSettler(
        ledger_api=mock.Mock(),
        crypto=mock.Mock(),
        chain_type=ChainType.LOCAL,
    )

    def _method(*args: Any, **kwargs: Any) -> Any:
        raise Exception("some error")

    with pytest.raises(ChainInteractionError):
        settler.transact(_method, contract="service_registry", kwargs={})  # type: ignore


def test_retriable_exception(capsys: Any) -> None:
    """Test retriable exception."""

    class _retriable_method:
        should_raise = True

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            if self.should_raise:
                self.should_raise = False
                raise Exception("wrong transaction nonce")
            return {}

    settler = TxSettler(
        ledger_api=mock.Mock(),
        crypto=mock.Mock(),
        chain_type=ChainType.LOCAL,
    )
    settler.transact(_retriable_method(), contract="service_registry", kwargs={})  # type: ignore
    captured = capsys.readouterr()
    assert "Error occured when interacting with chain" in captured.out


def test_repricing(capsys: Any) -> None:
    """Test retriable exception."""

    class _repricable_method:
        tx_dict = {
            "maxFeePerGas": 100,
            "maxPriorityFeePerGas": 100,
        }

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            return self.tx_dict

    def _raise_fee_too_low(tx_signed: Any, **kwargs: Any) -> None:
        if tx_signed["maxFeePerGas"] == 110:
            return
        raise Exception("FeeTooLow")

    def _reprice(old_price: Any, **kwargs: Any) -> Any:
        return {"maxFeePerGas": 110, "maxPriorityFeePerGas": 110}

    settler = TxSettler(
        ledger_api=mock.Mock(
            send_signed_transaction=_raise_fee_too_low,
            try_get_gas_pricing=_reprice,
        ),
        crypto=mock.Mock(
            sign_transaction=lambda transaction: transaction,
        ),
        chain_type=ChainType.LOCAL,
    )
    settler.transact(_repricable_method(), contract="service_registry", kwargs={})  # type: ignore
    captured = capsys.readouterr()
    assert "Repricing the transaction..." in captured.out


def test_repricing_bad_tx(capsys: Any) -> None:
    """Test retriable exception."""

    class _repricable_method:
        retries = 0

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            if self.retries == 0:
                self.retries += 1
                return {
                    "maxFeePerGas": 100,
                }
            return {
                "maxFeePerGas": 100,
                "maxPriorityFeePerGas": 100,
            }

    def _raise_fee_too_low(tx_signed: Any, **kwargs: Any) -> None:
        if tx_signed["maxFeePerGas"] == 110:
            return
        raise Exception("FeeTooLow")

    def _reprice(old_price: Any, **kwargs: Any) -> Any:
        return {"maxFeePerGas": 110, "maxPriorityFeePerGas": 110}

    settler = TxSettler(
        ledger_api=mock.Mock(
            send_signed_transaction=_raise_fee_too_low,
            try_get_gas_pricing=_reprice,
        ),
        crypto=mock.Mock(
            sign_transaction=lambda transaction: transaction,
        ),
        chain_type=ChainType.LOCAL,
    )
    settler.transact(_repricable_method(), contract="service_registry", kwargs={})  # type: ignore
    captured = capsys.readouterr()
    assert "Repricing the transaction..." in captured.out


def test_tx_not_found(capsys: Any) -> None:
    """Test retriable exception."""

    class _get_transaction_receipt:
        _should_raise = True

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            if self._should_raise:
                self._should_raise = False
                raise Exception("Transaction with hash HASH not found")
            return "receipt"

    settler = TxSettler(
        ledger_api=mock.Mock(
            api=mock.Mock(
                eth=mock.Mock(
                    get_transaction_receipt=_get_transaction_receipt(),
                )
            )
        ),
        crypto=mock.Mock(
            sign_transaction=lambda transaction: transaction,
        ),
        chain_type=ChainType.LOCAL,
    )
    settler.transact(mock.Mock(), contract="service_registry", kwargs={})  # type: ignore
    captured = capsys.readouterr()

    assert "Error getting transaction receipt" in captured.out
    assert "Transaction with hash HASH not found" in captured.out
    assert "Will retry in 3.0..." in captured.out


def test_already_known(capsys: Any) -> None:
    """Test AlreadyKnown exception."""

    class _method:
        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            return {}

    class _raise:
        _should_raise = True

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            if self._should_raise:
                self._should_raise = False
                raise Exception("AlreadyKnown")
            return {}

    settler = TxSettler(
        ledger_api=mock.Mock(
            send_signed_transaction=_raise(),
        ),
        crypto=mock.Mock(
            sign_transaction=lambda transaction: transaction,
        ),
        chain_type=ChainType.LOCAL,
    )
    with mock.patch.object(settler, "_already_known") as _mock:
        settler.transact(_method(), contract="service_registry", kwargs={})  # type: ignore
        _mock.assert_called_with("AlreadyKnown")


def test_should_rebuild_with_oldnonce(capsys: Any) -> None:
    """Test OldNonce exception."""

    class _method:
        call_count = 0

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            self.call_count += 1
            return {}

    class _raise:
        _should_raise = True

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            if self._should_raise:
                self._should_raise = False
                raise Exception("OldNonce")
            return {}

    builder = _method()
    settler = TxSettler(
        ledger_api=mock.Mock(
            send_signed_transaction=_raise(),
        ),
        crypto=mock.Mock(
            sign_transaction=lambda transaction: transaction,
        ),
        chain_type=ChainType.LOCAL,
    )
    settler.transact(builder, contract="service_registry", kwargs={})  # type: ignore
    assert (
        builder.call_count == 2
    )  # should be 2 since the build method will be called twice


def test_timeout() -> None:
    """Test transact."""

    def _waitable(*arg: Any, **kwargs: Any) -> None:
        return None

    settler = TxSettler(
        ledger_api=mock.Mock(
            api=mock.Mock(
                eth=mock.Mock(
                    get_transaction_receipt=_waitable,
                )
            )
        ),
        crypto=mock.Mock(),
        chain_type=ChainType.LOCAL,
        sleep=0.1,
        retries=1,
    )

    with pytest.raises(ChainTimeoutError):
        settler.transact(lambda **x: None, "service_registry", {})  # type: ignore



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test utils."""

from typing import Dict

import pytest
from aea.configurations.data_types import PublicId

from autonomy.chain.exceptions import DependencyError
from autonomy.chain.utils import get_ipfs_hash_from_uri, parse_public_id_from_metadata


def get_dummy_metadata(
    name: str = "valory/package",
    description: str = "description",
    code_uri: str = "code_uri",
    nft_hash: str = "nft_hash",
    version: str = "0.1.0",
) -> Dict:
    """Get dummy metadata."""

    return {
        "name": name,
        "description": description,
        "code_uri": code_uri,
        "image": nft_hash,
        "attributes": [{"trait_type": "version", "value": version}],
    }


def test_get_ipfs_hash_from_uri() -> None:
    """Test `get_ipfs_hash_from_uri` method"""

    assert get_ipfs_hash_from_uri("ipfs://SOME_HASH") == "SOME_HASH"


@pytest.mark.parametrize(
    "public_id_string",
    [
        "author/package_name",
        "component_type/author/name",
        "skill/author/name/0.1.0",
    ],
)
def test_parse_public_id_from_metadata(public_id_string: str) -> None:
    """Test verify `parse_public_id_from_metadata` method"""

    public_id = parse_public_id_from_metadata(
        id_string=public_id_string,
    )

    assert isinstance(public_id, PublicId)


def test_parse_public_id_from_metadata_fail() -> None:
    """Test verify `parse_public_id_from_metadata` method"""

    with pytest.raises(DependencyError, match="Invalid package name found `public_id`"):
        parse_public_id_from_metadata(
            id_string="public_id",
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the rounds for the offend_abci skill."""

import json
from enum import Enum
from typing import Dict, Optional, Set, Tuple

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
    DegenerateRound,
    OffenseStatusDecoder,
)
from packages.valory.skills.offend_abci.payloads import OffencesPayload


class Event(Enum):
    """Event enumeration for the Offend ABCI demo."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"


class OffendRound(CollectSameUntilThresholdRound):
    """A round in which the agents simulate an offence"""

    synchronized_data_class = BaseSynchronizedData
    payload_class = OffencesPayload
    extended_requirements: Tuple[str, ...] = tuple()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""

        if self.threshold_reached:
            self.context.state.round_sequence.offence_status = json.loads(
                self.most_voted_payload,
                cls=OffenseStatusDecoder,
            )
            return self.synchronized_data, Event.DONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY
        return None


class FinishedOffendRound(DegenerateRound):
    """Indicates the completion of `OffendRound`."""


class OffendAbciApp(AbciApp[Event]):
    """OffendAbciApp

    Initial round: OffendRound

    Initial states: {OffendRound}

    Transition states:
        0. OffendRound
            - done: 1.
            - no majority: 0.
            - round timeout: 0.
        1. FinishedOffendRound

    Final states: {FinishedOffendRound}

    Timeouts:
        round timeout: 30.0
    """

    initial_round_cls: AppState = OffendRound
    transition_function: AbciAppTransitionFunction = {
        OffendRound: {
            Event.DONE: FinishedOffendRound,
            Event.NO_MAJORITY: OffendRound,
            Event.ROUND_TIMEOUT: OffendRound,
        },
        FinishedOffendRound: {},
    }
    final_states: Set[AppState] = {FinishedOffendRound}
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
    }
    db_pre_conditions: Dict[AppState, Set[str]] = {OffendRound: set()}
    db_post_conditions: Dict[AppState, Set[str]] = {FinishedOffendRound: set()}
